# *****************************************************************************
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#


class _PerspectiveStateManager(object):
    """Internal state management class that controls when `_process` is called
    within the C++ Table internals.

    `_process()` notifies the engine to clear its queue of pending updates to be
    applied and reconciled. When Perspective runs within an event loop, we
    should use the loop whenever possible to batch calls to `_process()`.
    Callers that have access to an event loop implementation should set
    `queue_process` to their own function with `table_id` and `state_manager` as
    positional arguments.

    Override functions must be bound to this instance using `functools.partial`,
    i.e.: `functools.partial(queue_process_custom,
    state_manager=table._state_manager)

    The guarantee of `queue_process` is that `call_process` will be called,
    either on the next iteration of the event loop or before output is generated
    (through a serialization method, for example).

    Though each :obj:`~perspective.Table` contains a separate instance of the
    state manager, `TO_PROCESS`, which contains the `t_pool` objects for pending
    `_process` calls, is shared amongst all instances of the state manager.
    """
    TO_PROCESS = {}

    def __init__(self):
        """Create a new instance of the state manager, and enable the default behavior
        of calling `_process()` synchronously.

        New instances of :obj:`_PerspectiveStateManager` have no awareness of whether
        an event loop is present - the instance's `queue_process` method must be
        overridden.
        """
        self.queue_process = self._queue_process_immediate

    def set_process(self, pool, table_id):
        """Queue a `_process` call on the specified pool and table ID.

        Checks whether a `_process()` call has been queued already for the specified
        table, and calls `queue_process`, which MUST be implemented by the caller.

        Args:
            pool (:obj`libbinding.t_pool`): a `t_pool` object
            table_id (:obj`int`): a unique ID for the Table
        """
        if table_id not in _PerspectiveStateManager.TO_PROCESS:
            _PerspectiveStateManager.TO_PROCESS[table_id] = pool
            self.queue_process(table_id)

    def call_process(self, table_id):
        """Given a table_id, find the corresponding pool and call `process()`
        on it, which takes all the updates that have been queued, applies each
        update to the global Table state, and then clears the queue.

        Args:
            table_id (:obj`int`): The unique ID of the Table
        """
        pool = _PerspectiveStateManager.TO_PROCESS.get(table_id, None)
        if pool is not None:
            pool._process()
            self.remove_process(table_id)

    def remove_process(self, table_id):
        """Remove a pool from the execution cache, indicating that it should no
        longer be operated on.

        Args:
            table_id (:obj`int`): The unique ID of the Table
        """
        _PerspectiveStateManager.TO_PROCESS.pop(table_id, None)

    def _queue_process_immediate(self, table_id):
        """Immediately execute `call_process` on the pool as soon
        as `queue_process` is called.

        This is the default implementation of `queue_process` for environments
        without an event loop, meaning that calls to :obj:`~perspective.Table`'s
        `update()` method are immediately followed by a call to `_process`.

        Args:
            table_id (:obj`int`): The unique ID of the Table
        """
        self.call_process(table_id)



import math
import tributary.lazy as tl


class TestOps:
    def test_Negate(self):
        t = tl.Node(value=5)
        out = tl.Negate(t)
        assert out() == -5

    def test_Invert(self):
        t = tl.Node(value=5)
        out = tl.Invert(t)
        assert out() == 1 / 5

    def test_Add(self):
        t = tl.Node(value=5)
        out = tl.Add(t, t)
        assert out() == 10

    def test_Sub(self):
        t = tl.Node(value=5)
        out = tl.Sub(t, t)
        assert out() == 0

    def test_Mult(self):
        t = tl.Node(value=5)
        out = tl.Mult(t, t)
        assert out() == 25

    def test_Div(self):
        t = tl.Node(value=15)
        t2 = tl.Node(value=5)
        out = tl.Div(t, t2)
        assert out() == 3

    def test_Mod(self):
        t = tl.Node(value=3)
        t2 = tl.Node(value=2)
        out = tl.Mod(t, t2)
        assert out() == 1

    def test_Pow(self):
        t = tl.Node(value=3)
        t2 = tl.Node(value=2)
        out = tl.Pow(t, t2)
        assert out() == 9

    def test_Sum(self):
        t = tl.Node(value=3)
        out = tl.Sum(t, t, t, 3)
        assert out() == 12

    def test_Average(self):
        t = tl.Node(value=3)
        out = tl.Average(t, 2, 1)
        assert out() == 2

    def test_Not(self):
        t = tl.Node(value=2)
        out = tl.Not(t)
        assert out() == False

    def test_And(self):
        t = tl.Node(value=2)
        out = tl.And(t, t)
        assert out() == 2

    def test_Or(self):
        t = tl.Node(value=2)
        out = tl.Or(t, t)
        assert out() == 2

    def test_Equal(self):
        t = tl.Node(value=2)
        out = tl.Equal(t, 2)
        assert out()

    def test_NotEqual(self):
        t = tl.Node(value=2)
        out = tl.NotEqual(t, 1)
        assert out()

    def test_Lt(self):
        t = tl.Node(value=2)
        out = tl.Lt(t, 1)
        assert out() == False

    def test_Le(self):
        t = tl.Node(value=2)
        out = tl.Le(t, 2)
        assert out()

    def test_Gt(self):
        t = tl.Node(value=2)
        out = tl.Gt(t, 2)
        assert out() == False

    def test_Ge(self):
        t = tl.Node(value=2)
        out = tl.Ge(t, 1)
        assert out()

    def test_Log(self):
        t = tl.Node(value=2)
        out = tl.Log(t)
        assert out() == math.log(2)

    def test_Sin(self):
        t = tl.Node(value=2)
        out = tl.Sin(t)
        assert out() == math.sin(2)

    def test_Cos(self):
        t = tl.Node(value=2)
        out = tl.Cos(t)
        assert out() == math.cos(2)

    def test_Tan(self):
        t = tl.Node(value=2)
        out = tl.Tan(t)
        assert out() == math.tan(2)

    def test_Arcsin(self):
        t = tl.Node(value=2)
        out = tl.Arcsin(tl.Div(t, 3))
        assert out() == math.asin(2 / 3)

    def test_Arccos(self):
        t = tl.Node(value=2)
        out = tl.Arccos(tl.Div(t, 3))
        assert out() == math.acos(2 / 3)

    def test_Arctan(self):
        t = tl.Node(value=2)
        out = tl.Arctan(t)
        assert out() == math.atan(2)

    def test_Sqrt(self):
        t = tl.Node(value=9)
        out = tl.Sqrt(t)
        assert out() == 3.0

    def test_Abs(self):
        t = tl.Node(value=-2)
        out = tl.Abs(t)
        assert out() == 2

    def test_Exp(self):
        t = tl.Node(value=2)
        out = tl.Exp(t)
        assert out() == math.exp(2)

    def test_Erf(self):
        t = tl.Node(value=2)
        out = tl.Erf(t)
        assert out() == math.erf(2)

    def test_Floor(self):
        t = tl.Node(value=2.2)
        out = tl.Floor(t)
        assert out() == 2.

    def test_Ceil(self):
        t = tl.Node(value=2.2)
        out = tl.Ceil(t)
        assert out() == 3.

    def test_Round(self):
        t = tl.Node(value=2.222222)
        out = tl.Round(t, ndigits=2)
        assert out() == 2.22

    def test_Int(self):
        t = tl.Node(value=2.0)
        out = tl.Int(t)
        assert out() == 2

    def test_Float(self):
        t = tl.Node(value=2)
        out = tl.Float(t)
        assert out() == 2.0

    def test_Bool(self):
        t = tl.Node(value=2)
        out = tl.Bool(t)
        assert out()

    def test_Str(self):
        t = tl.Node(value=2)
        out = tl.Str(t)
        assert out() == '2'

    def test_Len(self):
        t = tl.Node(value=[1, 2, 3])
        out = tl.Len(t)
        assert out() == 3



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from datetime import date
import pandas as pd
import numpy as np
from perspective import Table, PerspectiveWidget
from ..common import superstore

DF = superstore(200)


class TestWidgetPandas:

    def test_widget_load_table_df(self):
        table = Table(DF)
        widget = PerspectiveWidget(table)
        assert widget.table.schema() == {'Country': str, 'index': int, 'Region': str, 'Category': str, 'City': str, 'Customer ID': str, 'Discount': float,
                                         'Order Date': date, 'Order ID': str, 'Postal Code': str, 'Product ID': str, 'Profit': float, 'Quantity': int,
                                         'Row ID': int, 'Sales': int, 'Segment': str, 'Ship Date': date, 'Ship Mode': str, 'State': str, 'Sub-Category': str}

        assert sorted(widget.columns) == sorted(['Category', 'City', 'Country', 'Customer ID', 'Discount', 'index', 'Order Date', 'Order ID', 'Postal Code',
                                                 'Product ID', 'Profit', 'Quantity', 'Region', 'Row ID', 'Sales', 'Segment', 'Ship Date',
                                                 'Ship Mode', 'State', 'Sub-Category'])
        view = widget.table.view()
        assert view.num_rows() == len(DF)
        assert view.num_columns() == len(DF.columns) + 1  # index column

    def test_widget_load_data_df(self):
        widget = PerspectiveWidget(DF)
        assert sorted(widget.columns) == sorted(['Category', 'City', 'Country', 'Customer ID', 'Discount', 'index', 'Order Date', 'Order ID', 'Postal Code',
                                                 'Product ID', 'Profit', 'Quantity', 'Region', 'Row ID', 'Sales', 'Segment', 'Ship Date',
                                                 'Ship Mode', 'State', 'Sub-Category'])
        view = widget.table.view()
        assert view.num_rows() == len(DF)
        assert view.num_columns() == 20

    def test_widget_load_series(self):
        series = pd.Series(DF["Profit"].values, name="profit")
        widget = PerspectiveWidget(series)
        assert widget.table.schema() == {'index': int, 'profit': float}

        assert sorted(widget.columns) == sorted(["index", "profit"])
        view = widget.table.view()
        assert view.num_rows() == len(DF)
        assert view.num_columns() == 2

    def test_widget_load_pivot_table(self):
        pivot_table = pd.pivot_table(DF, values='Discount', index=['Country', 'Region'], columns='Category')
        widget = PerspectiveWidget(pivot_table)
        assert widget.row_pivots == ['Country', 'Region']
        assert widget.column_pivots == []
        assert widget.columns == ['Financials', 'Industrials', 'Technology']
        # table should host flattened data
        view = widget.table.view()
        assert view.num_rows() == 5
        assert view.num_columns() == 6

    def test_widget_load_pivot_table_with_user_pivots(self):
        pivot_table = pd.pivot_table(DF, values='Discount', index=['Country', 'Region'], columns='Category')
        widget = PerspectiveWidget(pivot_table, row_pivots=["Category", "Segment"])
        assert widget.row_pivots == ['Category', 'Segment']
        assert widget.column_pivots == []
        assert widget.columns == ['Financials', 'Industrials', 'Technology']
        # table should host flattened data
        view = widget.table.view()
        assert view.num_rows() == 5
        assert view.num_columns() == 6

    def test_widget_load_row_pivots(self):
        df_pivoted = DF.set_index(['Country', 'Region'])
        widget = PerspectiveWidget(df_pivoted)
        assert widget.row_pivots == ['Country', 'Region']
        assert widget.column_pivots == []
        assert sorted(widget.columns) == sorted(['Category', 'City', 'Customer ID', 'Discount', 'Order Date', 'Order ID', 'Postal Code',
                                                 'Product ID', 'Profit', 'Quantity', 'Row ID', 'Sales', 'Segment', 'Ship Date',
                                                 'Ship Mode', 'State', 'Sub-Category'])
        assert widget.table.size() == 200
        view = widget.table.view()
        assert view.num_rows() == len(DF)
        assert view.num_columns() == len(DF.columns) + 1  # index column

    def test_widget_load_row_pivots_with_user_pivots(self):
        df_pivoted = DF.set_index(['Country', 'Region'])
        widget = PerspectiveWidget(df_pivoted, row_pivots=["Category", "Segment"])
        assert widget.row_pivots == ['Category', 'Segment']
        assert widget.column_pivots == []
        assert sorted(widget.columns) == sorted(['Category', 'City', 'Customer ID', 'Discount', 'Order Date', 'Order ID', 'Postal Code',
                                                 'Product ID', 'Profit', 'Quantity', 'Row ID', 'Sales', 'Segment', 'Ship Date',
                                                 'Ship Mode', 'State', 'Sub-Category'])
        assert widget.table.size() == 200
        view = widget.table.view()
        assert view.num_rows() == len(DF)
        assert view.num_columns() == len(DF.columns) + 1  # index column

    def test_widget_load_column_pivots(self):
        arrays = [np.array(['bar', 'bar', 'bar', 'bar', 'baz', 'baz', 'baz', 'baz', 'foo', 'foo', 'foo', 'foo', 'qux', 'qux', 'qux', 'qux']),
                  np.array(['one', 'one', 'two', 'two', 'one', 'one', 'two', 'two', 'one', 'one', 'two', 'two', 'one', 'one', 'two', 'two']),
                  np.array(['X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y'])]
        tuples = list(zip(*arrays))
        index = pd.MultiIndex.from_tuples(tuples, names=['first', 'second', 'third'])
        df_both = pd.DataFrame(np.random.randn(3, 16), index=['A', 'B', 'C'], columns=index)
        widget = PerspectiveWidget(df_both)
        assert widget.columns == [' ']
        assert widget.column_pivots == ['first', 'second', 'third']
        assert widget.row_pivots == ['index']

    def test_widget_load_column_pivots_preserve_user_settings(self):
        arrays = [np.array(['bar', 'bar', 'bar', 'bar', 'baz', 'baz', 'baz', 'baz', 'foo', 'foo', 'foo', 'foo', 'qux', 'qux', 'qux', 'qux']),
                  np.array(['one', 'one', 'two', 'two', 'one', 'one', 'two', 'two', 'one', 'one', 'two', 'two', 'one', 'one', 'two', 'two']),
                  np.array(['X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y'])]
        tuples = list(zip(*arrays))
        index = pd.MultiIndex.from_tuples(tuples, names=['first', 'second', 'third'])
        df_both = pd.DataFrame(np.random.randn(3, 16), index=['A', 'B', 'C'], columns=index)
        widget = PerspectiveWidget(df_both, columns=["first", "third"])
        assert widget.columns == ['first', "third"]
        assert widget.column_pivots == ['first', 'second', 'third']
        assert widget.row_pivots == ['index']



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import numpy as np
from math import trunc
from ._constants import COLUMN_SEPARATOR_STRING
from .libbinding import get_data_slice_zero, get_data_slice_one, get_data_slice_two, \
    get_from_data_slice_zero, get_from_data_slice_one, get_from_data_slice_two, \
    get_pkeys_from_data_slice_zero, get_pkeys_from_data_slice_one, get_pkeys_from_data_slice_two


def _mod(a, b):
    '''C-style modulo function'''
    if b == 0:
        # Javascript returns NaN in cases of division by 0; return -1 because
        # None would fail comparisons with other ints
        return float('nan')
    d = trunc(float(a) / b)
    return a - d * b


def to_format(options, view, output_format):
    view._table._state_manager.call_process(view._table._table.get_id())
    options, column_names, data_slice = _to_format_helper(view, options)

    if output_format == 'records':
        data = []
    elif output_format in ('dict', 'numpy'):
        data = {}
        if options["index"]:
            data['__INDEX__'] = []

    num_columns = len(view._config.get_columns())
    num_hidden = view._num_hidden_cols()

    for ridx in range(options["start_row"], options["end_row"]):
        row_path = data_slice.get_row_path(ridx) if options["has_row_path"] else []
        if options["leaves_only"] and len(row_path) < len(view._config.get_row_pivots()):
            continue

        if output_format == 'records':
            data.append({})

        for cidx in range(options["start_col"], options["end_col"]):
            name = column_names[cidx]

            if _mod((cidx - (1 if view._sides > 0 else 0)), (num_columns + num_hidden)) >= num_columns:
                # don't emit columns used for hidden sort
                continue
            elif cidx == options["start_col"] and view._sides > 0:
                if options["has_row_path"]:
                    paths = [path.to_string(False) for path in row_path]
                    paths.reverse()
                    if output_format == 'records':
                        data[-1]["__ROW_PATH__"] = paths
                    elif output_format in ('dict', 'numpy'):
                        if "__ROW_PATH__" not in data:
                            data["__ROW_PATH__"] = []
                        data["__ROW_PATH__"].append(paths)
            else:
                if output_format in ('dict', 'numpy') and (name not in data):
                    # TODO: push into C++ for numpy
                    data[name] = []
                if view._sides == 0:
                    value = get_from_data_slice_zero(data_slice, ridx, cidx)
                elif view._sides == 1:
                    value = get_from_data_slice_one(data_slice, ridx, cidx)
                else:
                    value = get_from_data_slice_two(data_slice, ridx, cidx)

                if output_format == 'records':
                    data[-1][name] = value
                else:
                    data[name].append(value)

        if options['index']:
            if view._sides == 0:
                pkeys = get_pkeys_from_data_slice_zero(data_slice, ridx, 0)
            elif view._sides == 1:
                pkeys = get_pkeys_from_data_slice_one(data_slice, ridx, 0)
            else:
                pkeys = get_pkeys_from_data_slice_two(data_slice, ridx, 0)

            if output_format == 'records':
                data[-1]['__INDEX__'] = []
                for pkey in pkeys:
                    data[-1]['__INDEX__'].append(pkey)
            elif output_format in ('dict', 'numpy'):
                # ensure that `__INDEX__` has the same number of rows as
                # returned dataset
                if len(pkeys) == 0:
                    data["__INDEX__"].append([])
                for pkey in pkeys:
                    data['__INDEX__'].append([pkey])

    if output_format in ('dict', 'numpy') and (not options["has_row_path"] and ("__ROW_PATH__" in data)):
        del data["__ROW_PATH__"]

    if output_format == 'numpy':
        for k, v in data.items():
            # TODO push into C++
            data[k] = np.array(v)

    return data


def _to_format_helper(view, options=None):
    '''Retrieves the data slice and column names in preparation for data
    serialization.
    '''
    options = options or {}
    opts = _parse_format_options(view, options)

    if view._sides == 0:
        data_slice = get_data_slice_zero(view._view, opts["start_row"], opts["end_row"], opts["start_col"], opts["end_col"])
    elif view._sides == 1:
        data_slice = get_data_slice_one(view._view, opts["start_row"], opts["end_row"], opts["start_col"], opts["end_col"])
    else:
        data_slice = get_data_slice_two(view._view, opts["start_row"], opts["end_row"], opts["start_col"], opts["end_col"])

    raw_names = data_slice.get_column_names()
    column_names = []

    for n in raw_names:
        column_names.append(COLUMN_SEPARATOR_STRING.join(
            [path.to_string(False) for path in n]))

    return opts, column_names, data_slice


def _parse_format_options(view, options):
    '''Given a user-provided options dictionary, extract the useful values.'''
    max_cols = view.num_columns() + (1 if view._sides > 0 else 0)
    return {
        "start_row": max(options.get("start_row", 0), 0),
        "end_row": min(options.get("end_row", view.num_rows()), view.num_rows()),
        "start_col": max(options.get("start_col", 0), 0),
        "end_col": min(options.get("end_col", max_cols) * (view._num_hidden_cols() + 1), max_cols),
        "index": options.get("index", False),
        "leaves_only": options.get("leaves_only", False),
        "has_row_path": view._sides > 0 and (not view._column_only)
    }



from .utils import _CALCULATIONS_GRAPHVIZSHAPE
from ..base import Node


def Count(node):
    '''Node to count inputs

    Args:
        node (Node): input stream
    '''
    def foo(val):
        ret._count += 1
        return ret._count

    ret = Node(foo=foo, foo_kwargs=None, name='Count', inputs=1, graphvizshape=_CALCULATIONS_GRAPHVIZSHAPE)
    ret._count = 0
    node._downstream.append((ret, 0))
    ret._upstream.append(node)
    return ret


def Max(node):
    '''Node to take rolling max of inputs

    Args:
        node (Node): input stream
    '''
    def foo(val):
        ret._max = max(ret._max, val) if ret._max is not None else val
        return ret._max

    ret = Node(foo=foo, foo_kwargs=None, name='Max', inputs=1, graphvizshape=_CALCULATIONS_GRAPHVIZSHAPE)
    ret._max = None

    node._downstream.append((ret, 0))
    ret._upstream.append(node)
    return ret


def Min(node):
    '''Node to take rolling min of inputs

    Args:
        node (Node): input stream
    '''
    def foo(val):
        ret._min = min(ret._min, val) if ret._min is not None else val
        return ret._min

    ret = Node(foo=foo, foo_kwargs=None, name='Min', inputs=1, graphvizshape=_CALCULATIONS_GRAPHVIZSHAPE)
    ret._min = None
    node._downstream.append((ret, 0))
    ret._upstream.append(node)
    return ret


def Sum(node):
    '''Node to take rolling sum inputs

    If stream type is iterable, will do += sum(input). If input
    stream type is not iterable, will do += input.

    Args:
        node (Node): input stream
    '''
    def foo(val):
        try:
            # iterable, sum with sum function
            iter(val)
            ret._sum += sum(val)
        except TypeError:
            # not iterable, sum by value
            ret._sum += val
        return ret._sum

    ret = Node(foo=foo, foo_kwargs=None, name='Sum', inputs=1, graphvizshape=_CALCULATIONS_GRAPHVIZSHAPE)
    ret._sum = 0

    node._downstream.append((ret, 0))
    ret._upstream.append(node)
    return ret


def Average(node):
    '''Node to take the running average

    If stream type is iterable, will do (average + sum(input))/(count+len(input)).
    If input stream type is not iterable, will do (average + input)/count
    '''
    def foo(val):
        try:
            # iterable, sum with sum function
            iter(val)
            ret._sum += sum(val)
            ret._count += len(val)
        except TypeError:
            # not iterable, sum by value
            ret._sum += val
            ret._count += 1
        return ret._sum / ret._count if ret._count > 0 else float('nan')

    ret = Node(foo=foo, foo_kwargs=None, name='Average', inputs=1, graphvizshape=_CALCULATIONS_GRAPHVIZSHAPE)
    ret._sum = 0
    ret._count = 0

    node._downstream.append((ret, 0))
    ret._upstream.append(node)
    return ret


Node.rollingCount = Count
Node.rollingMin = Min
Node.rollingMax = Max
Node.rollingSum = Sum
Node.rollingAverage = Average



import aiofiles
import json as JSON
from .input import Foo


class File(Foo):
    '''Open up a file and yield back lines in the file

    Args:
        filename (str): filename to read
        json (bool): load file line as json
    '''

    def __init__(self, filename, json=True):
        async def _file(filename=filename, json=json):
            async with aiofiles.open(filename) as f:
                async for line in f:
                    if json:
                        yield JSON.loads(line)
                    else:
                        yield line
        super().__init__(foo=_file)
        self._name = 'File'



import time
import random


class TestFunctional:
    def setup(self):
        pass
        # setup() before each test method

    def test_general(self):
        import tributary as t

        def foo1(on_data):
            x = 0
            while x < 5:
                on_data({'a': random.random(), 'b': random.randint(0, 1000), 'x': x})
                time.sleep(.01)
                x = x + 1

        def foo2(data, callback):
            callback([{'a': data['a'] * 1000, 'b': data['b'], 'c': 'AAPL', 'x': data['x']}])

        t.pipeline([foo1, foo2], ['on_data', 'callback'], on_data=lambda x: None)
        time.sleep(1)
        t.stop()



from temporalcache import interval, expire
from .base import Node


def Expire(node, second=None, minute=None, hour=None, day=None, day_of_week=None, week=None, month=None, maxsize=128):
    def _interval(node=node):
        return node()

    # make new node
    ret = node._gennode('Expire[{}-{}-{}-{}-{}-{}-{}-maxsize:{}](-'.format(second, minute, hour, day, day_of_week, week, month, maxsize) + node._name + ')', _interval, [node], node._trace)

    # stash original recompute
    ret._orig_recompute = ret._recompute

    # make recompute run on expire
    ret._recompute = expire(second=second, minute=minute, hour=hour, day=day, day_of_week=day_of_week, week=week, month=month, maxsize=maxsize)(ret._recompute)
    return ret


def Interval(node, seconds=0, minutes=0, hours=0, days=0, weeks=0, months=0, years=0, maxsize=128):
    def _interval(node=node):
        return node()

    # make new node
    ret = node._gennode('Interval[{}-{}-{}-{}-{}-{}-{}-maxsize:{}'.format(seconds, minutes, hours, days, weeks, months, years, maxsize) + node._name + ')', _interval, [node], node._trace)

    # stash original recompute
    ret._orig_recompute = ret._recompute

    # make recompute run on interval
    ret._recompute = interval(seconds=seconds, minutes=minutes, hours=hours, days=days, weeks=weeks, months=months, years=years, maxsize=maxsize)(ret._recompute)
    return ret


Node.expire = Expire
Node.interval = Interval



import tributary.lazy as tl


class TestOutput:
    def test_graphviz(self):
        def foo():
            return 1
        assert tl.GraphViz(tl.Node(callable=foo, always_dirty=True))

    def test_dagre(self):
        def foo():
            return 1

        assert tl.Dagre(tl.Node(callable=foo, always_dirty=True))



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import six
import numpy as np
from datetime import datetime

DATE_DTYPES = [np.dtype("datetime64[D]"), np.dtype("datetime64[W]"), np.dtype("datetime64[M]"), np.dtype("datetime64[Y]")]


def deconstruct_numpy(array):
    '''Given a numpy array, parse it and return the data as well as a numpy
    array of null indices.

    Args:
        array (numpy.array)

    Returns:
        `dict`: `array` is the original array, and `mask` is an array of
            booleans where `True` represents a nan/None value.
    '''
    mask = []

    is_object_or_string_dtype = np.issubdtype(array.dtype, np.str_) or\
        np.issubdtype(array.dtype, np.object_)

    if six.PY2:
        is_object_or_string_dtype = is_object_or_string_dtype or np.issubdtype(array.dtype, np.unicode_)

    is_datetime_dtype = np.issubdtype(array.dtype, np.datetime64) or\
        np.issubdtype(array.dtype, np.timedelta64)

    for i, item in enumerate(array):
        invalid = item is None

        if not is_object_or_string_dtype:
            if is_datetime_dtype:
                invalid = invalid or np.isnat(item)
            else:
                invalid = invalid or np.isnan(item)

        if invalid:
            mask.append(i)

    if array.dtype == bool or array.dtype == "?":
        # bool => byte
        array = array.astype("b", copy=False)
    elif np.issubdtype(array.dtype, np.datetime64):
        # treat days/weeks/months/years as datetime objects - avoid idiosyncracy
        # with days of month, etc.
        if array.dtype in DATE_DTYPES:
            array = array.astype(datetime)

        # cast datetimes to millisecond timestamps
        # because datetime64("nat") is a double, cast to float64 here - C++
        # handles the rest
        if array.dtype == np.dtype("datetime64[us]"):
            array = array.astype(np.float64, copy=False) / 1000
        elif array.dtype == np.dtype("datetime64[ns]"):
            array = array.astype(np.float64, copy=False) / 1000000
        elif array.dtype == np.dtype("datetime64[ms]"):
            array = array.astype(np.float64, copy=False)
        elif array.dtype == np.dtype("datetime64[s]"):
            array = array.astype(np.float64, copy=False) * 1000
        elif array.dtype == np.dtype("datetime64[m]"):
            array = array.astype(np.float64, copy=False) * 60000
        elif array.dtype == np.dtype("datetime64[h]"):
            array = array.astype(np.float64, copy=False) * 3600000
    elif np.issubdtype(array.dtype, np.timedelta64):
        array = array.astype(np.float64, copy=False)

    return {
            "array": array,
            "mask": mask
        }



import json as JSON
from socketIO_client_nexus import SocketIO as SIO
from urllib.parse import urlparse

from .input import Foo


class SocketIO(Foo):
    '''Connect to socketIO server and yield back results

    Args:
        url (str): url to connect to
        channel (str): socketio channel to connect through
        field (str): field to index result by
        sendinit (list): data to send on socketio connection open
        json (bool): load websocket data as json
        wrap (bool): wrap result in a list
        interval (int): socketio wai interval
    '''

    def __init__(self, url, channel='', field='', sendinit=None, json=False, wrap=False, interval=1):
        o = urlparse(url)
        socketIO = SIO(o.scheme + '://' + o.netloc, o.port)
        if sendinit:
            socketIO.emit(sendinit)

        async def _sio(url=url, channel=channel, field=field, json=json, wrap=wrap, interval=interval):
            while True:
                _data = []
                socketIO.on(channel, lambda data: _data.append(data))
                socketIO.wait(seconds=interval)
                for msg in _data:
                    # FIXME clear _data
                    if json:
                        msg = JSON.loads(msg)

                    if field:
                        msg = msg[field]

                    if wrap:
                        msg = [msg]

                    yield msg
        super().__init__(foo=_sio)
        self._name = 'SocketIO'



import time
import tributary.symbolic as ts
import tributary.streaming as tss
import sympy as sy
import numpy as np
import sympy as sy
from IPython.display import display, HTML
from sympy.stats import Normal as syNormal, cdf
sy.init_printing()


class TestConfig:
    def setup(self):
        pass
        # setup() before each test method

    def test_construct_lazy(self):
        # adapted from https://gist.github.com/raddy/bd0e977dc8437a4f8276
        spot, strike, vol, dte, rate, cp = sy.symbols('spot strike vol dte rate cp')

        T = dte / 260.
        N = syNormal('N', 0.0, 1.0)

        d1 = (sy.ln(spot / strike) + (0.5 * vol ** 2) * T) / (vol * sy.sqrt(T))
        d2 = d1 - vol * sy.sqrt(T)

        TimeValueExpr = sy.exp(-rate * T) * (cp * spot * cdf(N)(cp * d1) - cp * strike * cdf(N)(cp * d2))

        PriceClass = ts.construct_lazy(TimeValueExpr)

        price = PriceClass(spot=210.59, strike=205, vol=14.04, dte=4, rate=.2175, cp=-1)

        x = price.evaluate()()

        assert price.evaluate()() == x

        price.strike = 210

        assert x != price.evaluate()()

    def test_others(self):
        # adapted from https://gist.github.com/raddy/bd0e977dc8437a4f8276
        spot, strike, vol, dte, rate, cp = sy.symbols('spot strike vol dte rate cp')
        T = dte / 260.
        N = syNormal('N', 0.0, 1.0)
        d1 = (sy.ln(spot / strike) + (0.5 * vol ** 2) * T) / (vol * sy.sqrt(T))
        d2 = d1 - vol * sy.sqrt(T)
        TimeValueExpr = sy.exp(-rate * T) * (cp * spot * cdf(N)(cp * d1) - cp * strike * cdf(N)(cp * d2))
        PriceClass = ts.construct_lazy(TimeValueExpr)
        price = PriceClass(spot=210.59, strike=205, vol=14.04, dte=4, rate=.2175, cp=-1)
        price.evaluate()()
        ts.graphviz(TimeValueExpr)
        assert ts.traversal(TimeValueExpr)
        assert ts.symbols(TimeValueExpr)

    def test_parse(self):
        from sympy.parsing.sympy_parser import parse_expr
        assert parse_expr('x^2') == ts.parse_expression('x^2')

    def test_construct_streaming(self):
        # adapted from https://gist.github.com/raddy/bd0e977dc8437a4f8276
        # spot, strike, vol, days till expiry, interest rate, call or put (1,-1)
        spot, strike, vol, dte, rate, cp = sy.symbols('spot strike vol dte rate cp')

        T = dte / 260.
        N = syNormal('N', 0.0, 1.0)

        d1 = (sy.ln(spot / strike) + (0.5 * vol ** 2) * T) / (vol * sy.sqrt(T))
        d2 = d1 - vol * sy.sqrt(T)

        TimeValueExpr = sy.exp(-rate * T) * (cp * spot * cdf(N)(cp * d1) - cp * strike * cdf(N)(cp * d2))

        PriceClass = ts.construct_streaming(TimeValueExpr)

        def strikes():
            strike = 205
            while strike <= 220:
                yield strike
                strike += 2.5

        price = PriceClass(spot=tss.Const(210.59),
                           #    strike=tss.Print(tss.Const(205), text='strike'),
                           strike=tss.Foo(strikes, interval=1),
                           vol=tss.Const(14.04),
                           dte=tss.Const(4),
                           rate=tss.Const(.2175),
                           cp=tss.Const(-1))

        ret = tss.run(tss.Print(price._node))
        time.sleep(2)
        assert len(ret) == 7



# *****************************************************************************
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
import six
from io import StringIO
from datetime import date, datetime
import numpy as np
import pandas as pd
from perspective.table import Table
from ..common import superstore


class TestTablePandas(object):
    def test_empty_table(self):
        tbl = Table([])
        assert tbl.size() == 0
        assert tbl.schema() == {}

    def test_table_dataframe(self):
        d = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
        data = pd.DataFrame(d)
        tbl = Table(data)
        assert tbl.size() == 2
        assert tbl.schema() == {
            "index": int,
            "a": int,
            "b": int
        }
        assert tbl.view().to_records() == [
            {"a": 1, "b": 2, "index": 0},
            {"a": 3, "b": 4, "index": 1}
        ]

    def test_table_dataframe_column_order(self):
        d = [{"a": 1, "b": 2, "c": 3, "d": 4}, {"a": 3, "b": 4, "c": 5, "d": 6}]
        data = pd.DataFrame(d, columns=["b", "c", "a", "d"])
        tbl = Table(data)
        assert tbl.size() == 2
        assert tbl.columns() == ["index", "b", "c", "a", "d"]

    def test_table_dataframe_selective_column_order(self):
        d = [{"a": 1, "b": 2, "c": 3, "d": 4}, {"a": 3, "b": 4, "c": 5, "d": 6}]
        data = pd.DataFrame(d, columns=["b", "c", "a"])
        tbl = Table(data)
        assert tbl.size() == 2
        assert tbl.columns() == ["index", "b", "c", "a"]

    def test_table_dataframe_does_not_mutate(self):
        # make sure we don't mutate the dataframe that a user passes in
        data = pd.DataFrame({
            "a": np.array([None, 1, None, 2], dtype=object),
            "b": np.array([1.5, None, 2.5, None], dtype=object)
        })
        assert data["a"].tolist() == [None, 1, None, 2]
        assert data["b"].tolist() == [1.5, None, 2.5, None]

        tbl = Table(data)
        assert tbl.size() == 4
        assert tbl.schema() == {
            "index": int,
            "a": int,
            "b": float
        }

        assert data["a"].tolist() == [None, 1, None, 2]
        assert data["b"].tolist() == [1.5, None, 2.5, None]

    def test_table_date_series(self, util):
        data = util.make_series(freq="D")
        tbl = Table(data)
        assert tbl.size() == 10
        assert tbl.schema() == {
            "index": date,
            "0": float
        }
        assert tbl.view().to_dict()["index"] == [
            datetime(2000, 1, 1),
            datetime(2000, 1, 2),
            datetime(2000, 1, 3),
            datetime(2000, 1, 4),
            datetime(2000, 1, 5),
            datetime(2000, 1, 6),
            datetime(2000, 1, 7),
            datetime(2000, 1, 8),
            datetime(2000, 1, 9),
            datetime(2000, 1, 10)
        ]

    def test_table_time_series(self, util):
        data = util.make_series(freq="H")
        tbl = Table(data)
        assert tbl.size() == 10
        assert tbl.schema() == {
            "index": datetime,
            "0": float
        }
        assert tbl.view().to_dict()["index"] == [
            datetime(2000, 1, 1, 0, 0, 0),
            datetime(2000, 1, 1, 1, 0, 0),
            datetime(2000, 1, 1, 2, 0, 0),
            datetime(2000, 1, 1, 3, 0, 0),
            datetime(2000, 1, 1, 4, 0, 0),
            datetime(2000, 1, 1, 5, 0, 0),
            datetime(2000, 1, 1, 6, 0, 0),
            datetime(2000, 1, 1, 7, 0, 0),
            datetime(2000, 1, 1, 8, 0, 0),
            datetime(2000, 1, 1, 9, 0, 0)
        ]

    def test_table_dataframe_infer_date(self, util):
        data = util.make_dataframe(freq="M")

        tbl = Table(data)
        assert tbl.size() == 10
        assert tbl.schema() == {
            "index": date,
            "a": float,
            "b": float,
            "c": float,
            "d": float
        }

        assert tbl.view().to_dict()["index"] == [
            datetime(2000, 1, 31),
            datetime(2000, 2, 29),
            datetime(2000, 3, 31),
            datetime(2000, 4, 30),
            datetime(2000, 5, 31),
            datetime(2000, 6, 30),
            datetime(2000, 7, 31),
            datetime(2000, 8, 31),
            datetime(2000, 9, 30),
            datetime(2000, 10, 31)
        ]

    def test_table_dataframe_infer_time(self, util):
        data = util.make_dataframe(freq="H")

        tbl = Table(data)
        assert tbl.size() == 10
        assert tbl.schema() == {
            "index": datetime,
            "a": float,
            "b": float,
            "c": float,
            "d": float
        }

        assert tbl.view().to_dict()["index"] == [
            datetime(2000, 1, 1, 0, 0, 0),
            datetime(2000, 1, 1, 1, 0, 0),
            datetime(2000, 1, 1, 2, 0, 0),
            datetime(2000, 1, 1, 3, 0, 0),
            datetime(2000, 1, 1, 4, 0, 0),
            datetime(2000, 1, 1, 5, 0, 0),
            datetime(2000, 1, 1, 6, 0, 0),
            datetime(2000, 1, 1, 7, 0, 0),
            datetime(2000, 1, 1, 8, 0, 0),
            datetime(2000, 1, 1, 9, 0, 0)
        ]

    def test_table_dataframe_year_start_index(self, util):
        data = util.make_dataframe(freq="AS")

        tbl = Table(data)
        assert tbl.size() == 10
        assert tbl.schema() == {
            "index": date,
            "a": float,
            "b": float,
            "c": float,
            "d": float
        }

        assert tbl.view().to_dict()["index"] == [
            datetime(2000, 1, 1, 0, 0, 0),
            datetime(2001, 1, 1, 0, 0, 0),
            datetime(2002, 1, 1, 0, 0, 0),
            datetime(2003, 1, 1, 0, 0, 0),
            datetime(2004, 1, 1, 0, 0, 0),
            datetime(2005, 1, 1, 0, 0, 0),
            datetime(2006, 1, 1, 0, 0, 0),
            datetime(2007, 1, 1, 0, 0, 0),
            datetime(2008, 1, 1, 0, 0, 0),
            datetime(2009, 1, 1, 0, 0, 0)
        ]

    def test_table_dataframe_quarter_index(self, util):
        data = util.make_dataframe(size=4, freq="Q")

        tbl = Table(data)
        assert tbl.size() == 4
        assert tbl.schema() == {
            "index": date,
            "a": float,
            "b": float,
            "c": float,
            "d": float
        }

        assert tbl.view().to_dict()["index"] == [
            datetime(2000, 3, 31, 0, 0, 0),
            datetime(2000, 6, 30, 0, 0, 0),
            datetime(2000, 9, 30, 0, 0, 0),
            datetime(2000, 12, 31, 0, 0, 0)
        ]

    def test_table_dataframe_minute_index(self, util):
        data = util.make_dataframe(size=5, freq="min")

        tbl = Table(data)
        assert tbl.size() == 5
        assert tbl.schema() == {
            "index": datetime,
            "a": float,
            "b": float,
            "c": float,
            "d": float
        }

        assert tbl.view().to_dict()["index"] == [
            datetime(2000, 1, 1, 0, 0),
            datetime(2000, 1, 1, 0, 1),
            datetime(2000, 1, 1, 0, 2),
            datetime(2000, 1, 1, 0, 3),
            datetime(2000, 1, 1, 0, 4)
        ]

    def test_table_pandas_periodindex(self, util):
        df = util.make_period_dataframe(30)
        tbl = Table(df)

        assert tbl.size() == 30
        assert tbl.schema() == {
            "index": date,
            "a": float,
            "b": float,
            "c": float,
            "d": float
        }

        assert tbl.view().to_dict()["index"][:5] == [
            datetime(2000, 1, 1),
            datetime(2000, 2, 1),
            datetime(2000, 3, 1),
            datetime(2000, 4, 1),
            datetime(2000, 5, 1)
        ]

    def test_table_pandas_period(self):
        df = pd.DataFrame({"a": [pd.Period("1Q2019"), pd.Period("2Q2019"), pd.Period("3Q2019"), pd.Period("4Q2019")]})
        tbl = Table(df)
        assert tbl.size() == 4
        assert tbl.schema() == {
            "index": int,
            "a": datetime
        }
        assert tbl.view().to_dict()["a"] == [
            datetime(2019, 1, 1),
            datetime(2019, 4, 1),
            datetime(2019, 7, 1),
            datetime(2019, 10, 1),
        ]

    def test_table_pandas_from_schema_int(self):
        data = [None, 1, None, 2, None, 3, 4]
        df = pd.DataFrame({
            "a": data
        })
        table = Table({
            "a": int
        })
        table.update(df)
        assert table.view().to_dict()["a"] == data

    def test_table_pandas_from_schema_long(self):
        if six.PY2:
            data = [None, 1, None, 2, None, 3, 4]
            df = pd.DataFrame({
                "a": data
            })
            table = Table({
                "a": long  # noqa: F821
            })
            table.update(df)
            assert table.view().to_dict()["a"] == data

    def test_table_pandas_from_schema_bool(self):
        data = [True, False, True, False]
        df = pd.DataFrame({
            "a": data
        })
        table = Table({
            "a": bool
        })
        table.update(df)
        assert table.view().to_dict()["a"] == data

    def test_table_pandas_from_schema_bool_str(self):
        data = ["True", "False", "True", "False"]
        df = pd.DataFrame({
            "a": data
        })
        table = Table({
            "a": bool
        })
        table.update(df)
        assert table.view().to_dict()["a"] == [
            True, False, True, False
        ]

    def test_table_pandas_from_schema_float(self):
        data = [None, 1.5, None, 2.5, None, 3.5, 4.5]
        df = pd.DataFrame({
            "a": data
        })
        table = Table({
            "a": float
        })
        table.update(df)
        assert table.view().to_dict()["a"] == data

    def test_table_pandas_from_schema_float_all_nan(self):
        data = [np.nan, np.nan, np.nan, np.nan]
        df = pd.DataFrame({
            "a": data
        })
        table = Table({
            "a": float
        })
        table.update(df)
        assert table.view().to_dict()["a"] == [None, None, None, None]

    def test_table_pandas_from_schema_float_to_int(self):
        data = [None, 1.5, None, 2.5, None, 3.5, 4.5]
        df = pd.DataFrame({
            "a": data
        })
        table = Table({
            "a": int
        })
        table.update(df)
        # truncates decimal
        assert table.view().to_dict()["a"] == [None, 1, None, 2, None, 3, 4]

    def test_table_pandas_from_schema_int_to_float(self):
        data = [None, 1, None, 2, None, 3, 4]
        df = pd.DataFrame({
            "a": data
        })
        table = Table({
            "a": float
        })
        table.update(df)
        assert table.view().to_dict()["a"] == [None, 1.0, None, 2.0, None, 3.0, 4.0]

    def test_table_pandas_from_schema_date(self):
        data = [date(2019, 8, 15), None, date(2019, 8, 16)]
        df = pd.DataFrame({
            "a": data
        })
        table = Table({
            "a": date
        })
        table.update(df)
        assert table.view().to_dict()["a"] == [datetime(2019, 8, 15), None, datetime(2019, 8, 16)]

    def test_table_pandas_from_schema_datetime(self):
        data = [datetime(2019, 7, 11, 12, 30, 5), None, datetime(2019, 7, 11, 13, 30, 5), None]
        df = pd.DataFrame({
            "a": data
        })
        table = Table({
            "a": datetime
        })
        table.update(df)
        assert table.view().to_dict()["a"] == data

    def test_table_pandas_from_schema_datetime_timestamp_s(self, util):
        data = [util.to_timestamp(datetime(2019, 7, 11, 12, 30, 5)), np.nan, util.to_timestamp(datetime(2019, 7, 11, 13, 30, 5)), np.nan]
        df = pd.DataFrame({
            "a": data
        })
        table = Table({
            "a": datetime
        })
        table.update(df)
        assert table.view().to_dict()["a"] == [
            datetime(2019, 7, 11, 12, 30, 5),
            None,
            datetime(2019, 7, 11, 13, 30, 5),
            None
        ]

    def test_table_pandas_from_schema_datetime_timestamp_ms(self, util):
        data = [
            util.to_timestamp(datetime(2019, 7, 11, 12, 30, 5)) * 1000,
            np.nan,
            util.to_timestamp(datetime(2019, 7, 11, 13, 30, 5)) * 1000,
            np.nan
        ]

        df = pd.DataFrame({
            "a": data
        })
        table = Table({
            "a": datetime
        })
        table.update(df)
        assert table.view().to_dict()["a"] == [
            datetime(2019, 7, 11, 12, 30, 5),
            None,
            datetime(2019, 7, 11, 13, 30, 5),
            None
        ]

    def test_table_pandas_from_schema_str(self):
        data = ["a", None, "b", None, "c"]
        df = pd.DataFrame({
            "a": data
        })
        table = Table({
            "a": str
        })
        table.update(df)
        assert table.view().to_dict()["a"] == data

    def test_table_pandas_none(self):
        data = [None, None, None]
        df = pd.DataFrame({
            "a": data
        })
        table = Table(df)
        assert table.view().to_dict()["a"] == data

    def test_table_pandas_symmetric_table(self):
        # make sure that updates are symmetric to table creation
        df = pd.DataFrame({
            "a": [1, 2, 3, 4],
            "b": [1.5, 2.5, 3.5, 4.5]
        })
        t1 = Table(df)
        t2 = Table({
            "a": int,
            "b": float
        })
        t2.update(df)
        assert t1.view().to_dict() == {
            "index": [0, 1, 2, 3],
            "a": [1, 2, 3, 4],
            "b": [1.5, 2.5, 3.5, 4.5]
        }

    def test_table_pandas_symmetric_stacked_updates(self):
        # make sure that updates are symmetric to table creation
        df = pd.DataFrame({
            "a": [1, 2, 3, 4],
            "b": [1.5, 2.5, 3.5, 4.5]
        })

        t1 = Table(df)
        t1.update(df)

        t2 = Table({
            "a": int,
            "b": float
        })
        t2.update(df)
        t2.update(df)

        assert t1.view().to_dict() == {
            "index": [0, 1, 2, 3, 0, 1, 2, 3],
            "a": [1, 2, 3, 4, 1, 2, 3, 4],
            "b": [1.5, 2.5, 3.5, 4.5, 1.5, 2.5, 3.5, 4.5]
        }

    def test_table_pandas_transitive(self):
        # serialized output -> table -> serialized output
        df = pd.DataFrame({
            "a": [1, 2, 3, 4],
            "b": [1.5, 2.5, 3.5, 4.5],
            "c": [np.nan, np.nan, "abc", np.nan],
            "d": [None, True, None, False],
            "e": [float("nan"), datetime(2019, 7, 11, 12, 30), float("nan"), datetime(2019, 7, 11, 12, 30)]
        })

        t1 = Table(df)
        out1 = t1.view().to_df()

        t2 = Table(out1)

        assert t1.schema() == t2.schema()

        out2 = t2.view().to_dict()

        assert t1.view().to_dict() == out2

    # dtype=object should have correct inferred types

    def test_table_pandas_object_to_int(self):
        df = pd.DataFrame({
            "a": np.array([1, 2, None, 2, None, 3, 4], dtype=object)
        })
        table = Table(df)
        assert table.schema() == {
            "index": int,
            "a": int
        }
        assert table.view().to_dict()["a"] == [1, 2, None, 2, None, 3, 4]

    def test_table_pandas_object_to_float(self):
        df = pd.DataFrame({
            "a": np.array([None, 1, None, 2, None, 3, 4], dtype=object)
        })
        table = Table(df)
        assert table.schema() == {
            "index": int,
            "a": int
        }
        assert table.view().to_dict()["a"] == [None, 1.0, None, 2.0, None, 3.0, 4.0]

    def test_table_pandas_object_to_bool(self):
        df = pd.DataFrame({
            "a": np.array([True, False, True, False, True, False], dtype=object)
        })
        table = Table(df)
        assert table.schema() == {
            "index": int,
            "a": bool
        }
        assert table.view().to_dict()["a"] == [True, False, True, False, True, False]

    def test_table_pandas_object_to_date(self):
        df = pd.DataFrame({
            "a": np.array([date(2019, 7, 11), date(2019, 7, 12), None], dtype=object)
        })
        table = Table(df)
        assert table.schema() == {
            "index": int,
            "a": date
        }
        assert table.view().to_dict()["a"] == [datetime(2019, 7, 11), datetime(2019, 7, 12), None]

    def test_table_pandas_object_to_datetime(self):
        df = pd.DataFrame({
            "a": np.array([datetime(2019, 7, 11, 1, 2, 3), datetime(2019, 7, 12, 1, 2, 3), None], dtype=object)
        })
        table = Table(df)
        assert table.schema() == {
            "index": int,
            "a": datetime
        }
        assert table.view().to_dict()["a"] == [datetime(2019, 7, 11, 1, 2, 3), datetime(2019, 7, 12, 1, 2, 3), None]

    def test_table_pandas_object_to_str(self):
        df = pd.DataFrame({
            "a": np.array(["abc", "def", None, "ghi"], dtype=object)
        })
        table = Table(df)
        assert table.schema() == {
            "index": int,
            "a": str
        }
        assert table.view().to_dict()["a"] == ["abc", "def", None, "ghi"]

    # Type matching

    def test_table_pandas_update_float_schema_with_int(self):
        df = pd.DataFrame({
            "a": [1.5, 2.5, 3.5, 4.5],
            "b": [1, 2, 3, 4]
        })

        table = Table({
            "a": float,
            "b": float
        })

        table.update(df)

        assert table.view().to_dict() == {
            "a": [1.5, 2.5, 3.5, 4.5],
            "b": [1.0, 2.0, 3.0, 4.0]
        }

    def test_table_pandas_update_int32_with_int64(self):
        df = pd.DataFrame({
            "a": [1, 2, 3, 4]
        })

        table = Table({
            "a": [1, 2, 3, 4]
        })

        table.update(df)

        assert table.view().to_dict() == {
            "a": [1, 2, 3, 4, 1, 2, 3, 4]
        }

    def test_table_pandas_update_int64_with_float(self):
        df = pd.DataFrame({
            "a": [1.5, 2.5, 3.5, 4.5]
        })

        table = Table(pd.DataFrame({
            "a": [1, 2, 3, 4]
        }))

        table.update(df)

        assert table.view().to_dict()["a"] == [1, 2, 3, 4, 1, 2, 3, 4]

    def test_table_pandas_update_date_schema_with_datetime(self):
        df = pd.DataFrame({
            "a": np.array([date(2019, 7, 11)])
        })

        table = Table({
            "a": date
        })

        table.update(df)

        assert table.schema() == {
            "a": date
        }

        assert table.view().to_dict() == {
            "a": [datetime(2019, 7, 11)]
        }

    def test_table_pandas_update_datetime_schema_with_date(self):
        df = pd.DataFrame({
            "a": np.array([date(2019, 7, 11)])
        })

        table = Table({
            "a": datetime
        })

        table.update(df)

        assert table.schema() == {
            "a": datetime
        }

        assert table.view().to_dict() == {
            "a": [datetime(2019, 7, 11, 0, 0)]
        }

    # Timestamps

    def test_table_pandas_timestamp_to_datetime(self):
        data = [pd.Timestamp(2019, 7, 11, 12, 30, 5), None, pd.Timestamp(2019, 7, 11, 13, 30, 5), None]
        df = pd.DataFrame({
            "a": data
        })
        table = Table(df)
        assert table.view().to_dict()["a"] == [datetime(2019, 7, 11, 12, 30, 5), None, datetime(2019, 7, 11, 13, 30, 5), None]

    def test_table_pandas_timestamp_explicit_dtype(self):
        data = [pd.Timestamp(2019, 7, 11, 12, 30, 5), None, pd.Timestamp(2019, 7, 11, 13, 30, 5), None]
        df = pd.DataFrame({
            "a": np.array(data, dtype="datetime64[ns]")
        })
        table = Table(df)
        assert table.view().to_dict()["a"] == [datetime(2019, 7, 11, 12, 30, 5), None, datetime(2019, 7, 11, 13, 30, 5), None]

    def test_table_pandas_update_datetime_with_timestamp(self):
        data = [datetime(2019, 7, 11, 12, 30, 5), None, datetime(2019, 7, 11, 13, 30, 5), None]
        df = pd.DataFrame({
            "a": data
        })
        df2 = pd.DataFrame({
            "a": [pd.Timestamp(2019, 7, 11, 12, 30, 5), None, pd.Timestamp(2019, 7, 11, 13, 30, 5), None]
        })
        table = Table(df)
        table.update(df2)
        assert table.view().to_dict()["a"] == [datetime(2019, 7, 11, 12, 30, 5), None, datetime(2019, 7, 11, 13, 30, 5), None,
                                               datetime(2019, 7, 11, 12, 30, 5), None, datetime(2019, 7, 11, 13, 30, 5), None]

    # NaN/NaT reading

    def test_table_pandas_nan(self):
        data = [np.nan, np.nan, np.nan, np.nan]
        df = pd.DataFrame({
            "a": data
        })
        table = Table(df)
        assert table.view().to_dict()["a"] == [None, None, None, None]

    def test_table_pandas_int_nan(self):
        data = [np.nan, 1, np.nan, 2]
        df = pd.DataFrame({
            "a": data
        })
        table = Table(df)
        assert table.view().to_dict()["a"] == [None, 1, None, 2]

    def test_table_pandas_float_nan(self):
        data = [np.nan, 1.5, np.nan, 2.5]
        df = pd.DataFrame({
            "a": data
        })
        table = Table(df)
        assert table.view().to_dict()["a"] == [None, 1.5, None, 2.5]

    def test_table_read_nan_int_col(self):
        data = pd.DataFrame({"str": ["abc", float("nan"), "def"], "int": [np.nan, 1, 2]})
        tbl = Table(data)
        assert tbl.schema() == {
            "index": int,
            "str": str,
            "int": float  # np.nan is float type - ints convert to floats when filled in
        }
        assert tbl.size() == 3
        assert tbl.view().to_dict() == {
            "index": [0, 1, 2],
            "str": ["abc", None, "def"],
            "int": [None, 1.0, 2.0]
        }

    def test_table_read_nan_float_col(self):
        data = pd.DataFrame({"str": [float("nan"), "abc", float("nan")], "float": [np.nan, 1.5, 2.5]})
        tbl = Table(data)
        assert tbl.schema() == {
            "index": int,
            "str": str,
            "float": float  # can only promote to string or float
        }
        assert tbl.size() == 3
        assert tbl.view().to_dict() == {
            "index": [0, 1, 2],
            "str": [None, "abc", None],
            "float": [None, 1.5, 2.5]
        }

    def test_table_read_nan_bool_col(self):
        data = pd.DataFrame({"bool": [np.nan, True, np.nan], "bool2": [False, np.nan, True]})
        tbl = Table(data)
        # if np.nan begins a column, it is inferred as float and then can be promoted. if np.nan is in the values (but not at start), the column type is whatever is inferred.
        assert tbl.schema() == {
            "index": int,
            "bool": str,
            "bool2": bool
        }
        assert tbl.size() == 3
        # np.nans are always serialized as None
        assert tbl.view().to_dict() == {
            "index": [0, 1, 2],
            "bool": [None, "True", None],
            "bool2": [False, None, True]
        }

    def test_table_read_nan_date_col(self):
        data = pd.DataFrame({"str": ["abc", "def"], "date": [float("nan"), date(2019, 7, 11)]})
        tbl = Table(data)
        assert tbl.schema() == {
            "index": int,
            "str": str,
            "date": str  # can only promote to string or float
        }
        assert tbl.size() == 2
        assert tbl.view().to_dict() == {
            "index": [0, 1],
            "str": ["abc", "def"],
            "date": [None, '2019-07-11']
        }

    def test_table_read_nan_datetime_col(self):
        data = pd.DataFrame({"str": ["abc", "def"], "datetime": [float("nan"), datetime(2019, 7, 11, 11, 0)]})
        tbl = Table(data)
        assert tbl.schema() == {
            "index": int,
            "str": str,
            "datetime": datetime  # can only promote to string or float
        }
        assert tbl.size() == 2
        assert tbl.view().to_dict() == {
            "index": [0, 1],
            "str": ["abc", "def"],
            "datetime": [None, datetime(2019, 7, 11, 11, 0)]
        }

    def test_table_read_nat_datetime_col(self):
        data = pd.DataFrame({"str": ["abc", "def"], "datetime": ["NaT", datetime(2019, 7, 11, 11, 0)]})
        tbl = Table(data)
        assert tbl.schema() == {
            "index": int,
            "str": str,
            "datetime": datetime  # can only promote to string or float
        }
        assert tbl.size() == 2
        assert tbl.view().to_dict() == {
            "index": [0, 1],
            "str": ["abc", "def"],
            "datetime": [None, datetime(2019, 7, 11, 11, 0)]
        }

    def test_table_read_nan_datetime_as_date_col(self):
        data = pd.DataFrame({"str": ["abc", "def"], "datetime": [float("nan"), datetime(2019, 7, 11)]})
        tbl = Table(data)
        assert tbl.schema() == {
            "index": int,
            "str": str,
            "datetime": datetime  # can only promote to string or float
        }
        assert tbl.size() == 2
        assert tbl.view().to_dict() == {
            "index": [0, 1],
            "str": ["abc", "def"],
            "datetime": [None, datetime(2019, 7, 11)]
        }

    def test_table_read_nan_datetime_no_seconds(self):
        data = pd.DataFrame({"str": ["abc", "def"], "datetime": [float("nan"), datetime(2019, 7, 11, 11, 0)]})
        tbl = Table(data)
        assert tbl.schema() == {
            "index": int,
            "str": str,
            "datetime": datetime  # can only promote to string or float
        }
        assert tbl.size() == 2
        assert tbl.view().to_dict() == {
            "index": [0, 1],
            "str": ["abc", "def"],
            "datetime": [None, datetime(2019, 7, 11, 11, 0)]
        }

    def test_table_read_nan_datetime_milliseconds(self):
        data = pd.DataFrame({"str": ["abc", "def"], "datetime": [np.nan, datetime(2019, 7, 11, 10, 30, 55)]})
        tbl = Table(data)
        assert tbl.schema() == {
            "index": int,
            "str": str,
            "datetime": datetime  # can only promote to string or float
        }
        assert tbl.size() == 2
        assert tbl.view().to_dict() == {
            "index": [0, 1],
            "str": ["abc", "def"],
            "datetime": [None, datetime(2019, 7, 11, 10, 30, 55)]
        }

    def test_table_correct_csv_nan_end(self):
        s = "str,int\n,1\n,2\nabc,3"
        if six.PY2:
            s = unicode(s)
        csv = StringIO(s)
        data = pd.read_csv(csv)
        tbl = Table(data)
        assert tbl.schema() == {
            "index": int,
            "str": str,
            "int": int
        }
        assert tbl.size() == 3
        assert tbl.view().to_dict() == {
            "index": [0, 1, 2],
            "str": [None, None, "abc"],
            "int": [1, 2, 3]
        }

    def test_table_correct_csv_nan_intermittent(self):
        s = "str,float\nabc,\n,2\nghi,"
        if six.PY2:
            s = unicode(s)
        csv = StringIO(s)
        data = pd.read_csv(csv)
        tbl = Table(data)
        assert tbl.schema() == {
            "index": int,
            "str": str,
            "float": float
        }
        assert tbl.size() == 3
        assert tbl.view().to_dict() == {
            "index": [0, 1, 2],
            "str": ["abc", None, "ghi"],
            "float": [None, 2, None]
        }

    def test_table_series(self):
        import pandas as pd
        data = pd.Series([1, 2, 3], name="a")
        tbl = Table(data)
        assert tbl.size() == 3

    def test_table_indexed_series(self):
        import pandas as pd
        data = pd.Series([1, 2, 3], index=["a", "b", "c"], name="a")
        tbl = Table(data)
        assert tbl.schema() == {
            "index": str,
            "a": int
        }
        assert tbl.size() == 3

    def test_rowpivots(self):
        df = superstore(100)
        df_pivoted = df.set_index(['Country', 'Region'])
        table = Table(df_pivoted)
        columns = table.columns()
        assert table.size() == 100
        assert "Country" in columns
        assert "Region" in columns

    def test_pivottable(self):
        df = superstore()
        pt = pd.pivot_table(df, values='Discount', index=['Country', 'Region'], columns='Category')
        table = Table(pt)
        columns = table.columns()
        assert "Country" in columns
        assert "Region" in columns

    def test_colpivots(self):
        arrays = [np.array(['bar', 'bar', 'bar', 'bar', 'baz', 'baz', 'baz', 'baz', 'foo', 'foo', 'foo', 'foo', 'qux', 'qux', 'qux', 'qux']),
                  np.array(['one', 'one', 'two', 'two', 'one', 'one', 'two', 'two', 'one', 'one', 'two', 'two', 'one', 'one', 'two', 'two']),
                  np.array(['X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y'])]
        tuples = list(zip(*arrays))
        index = pd.MultiIndex.from_tuples(tuples, names=['first', 'second', 'third'])

        df_both = pd.DataFrame(np.random.randn(3, 16), index=['A', 'B', 'C'], columns=index)
        table = Table(df_both)
        assert table.size() == 48



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from enum import Enum


class Sort(Enum):
    '''The sort directions available for use in Perspective. Pass these into
    the `sort` argument of `PerspectiveWidget`.

    Examples:
        >>> widget = PerspectiveWidget(data, sort=[["a", Sort.DESC]])
    '''
    ASC = 'asc'
    ASC_ABS = 'asc abs'
    COL_ASC = 'col asc'
    COL_ASC_ABS = 'col asc abs'
    COL_DESC = 'col desc'
    COL_DESC_ABS = 'col desc abs'
    DESC = 'desc'
    DESC_ABS = 'desc abs'
    NONE = 'none'

    @staticmethod
    def options():
        return list(c.value for c in Sort)



import aiohttp
import json as JSON
import time
from .input import Foo


class HTTP(Foo):
    '''Connect to url and yield results

    Args:
        url (str): url to connect to
        interval (int): interval to re-query
        repeat (int): number of times to request
        json (bool): load http content data as json
        wrap (bool): wrap result in a list
        field (str): field to index result by
        proxies (list): list of URL proxies to pass to requests.get
        cookies (list): list of cookies to pass to requests.get
    '''

    def __init__(self, url, interval=1, repeat=1, json=False, wrap=False, field=None, proxies=None, cookies=None):
        async def _req(url=url, interval=interval, repeat=repeat, json=json, wrap=wrap, field=field, proxies=proxies, cookies=cookies):
            count = 0
            while count < repeat:
                async with aiohttp.ClientSession() as session:
                    async with session.get(url, cookies=cookies, proxy=proxies) as response:
                        msg = await response.text()

                        if msg is None or response.status != 200:
                            break

                        if json:
                            msg = JSON.loads(msg)

                        if field:
                            msg = msg[field]

                        if wrap:
                            msg = [msg]

                        yield msg

                        if interval:
                            time.sleep(interval)
                        if repeat >= 0:
                            count += 1

        super().__init__(foo=_req)
        self._name = 'Http'



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import six
import numpy
import pandas
import json
from datetime import date, datetime
from functools import partial
from ipywidgets import Widget
from traitlets import observe, Unicode
from ..core.data import deconstruct_pandas
from ..core.exception import PerspectiveError
from ..libpsp import is_libpsp
from ..viewer import PerspectiveViewer
from ..core._version import major_minor_version


def _type_to_string(t):
    '''Convert a type object to a string representing a Perspective-supported
    type.  Redefine here as we can't have any dependencies on libbinding in
    client mode.
    '''
    if t in six.integer_types:
        return "integer"
    elif t is float:
        return "float"
    elif t is bool:
        return "boolean"
    elif t is date:
        return "date"
    elif t is datetime:
        return "datetime"
    elif t is six.binary_type or t is six.text_type:
        return "string"
    else:
        raise PerspectiveError(
            "Unsupported type `{0}` in schema - Perspective supports `int`, `float`, `bool`, `date`, `datetime`, and `str` (or `unicode`).".format(str(t)))


def _serialize(data):
    # Attempt to serialize data and pass it to the front-end as JSON
    if isinstance(data, list):
        return data
    elif isinstance(data, dict):
        for v in data.values():
            # serialize schema values to string
            if isinstance(v, type):
                return {k: _type_to_string(data[k]) for k in data}
            elif isinstance(v, numpy.ndarray):
                return {k: data[k].tolist() for k in data}
            else:
                return data
    elif isinstance(data, numpy.ndarray):
        # structured or record array
        if not isinstance(data.dtype.names, tuple):
            raise NotImplementedError("Data should be dict of numpy.ndarray or a structured array.")
        columns = [data[col].tolist() for col in data.dtype.names]
        return dict(zip(data.dtype.names, columns))
    elif isinstance(data, pandas.DataFrame) or isinstance(data, pandas.Series):
        # Take flattened dataframe and make it serializable
        d = {}
        for name in data.columns:
            column = data[name]
            values = column.values
            # Timezone-aware datetime64 dtypes throw an exception when using
            # `numpy.issubdtype` - match strings here instead.
            str_dtype = str(column.dtype)
            if "datetime64" in str_dtype:
                # Convert all datetimes to string for serializing
                values = numpy.datetime_as_string(column.values, unit="ms")
            d[name] = values.tolist()
        return d
    else:
        raise NotImplementedError("Cannot serialize a dataset of `{0}`.".format(str(type(data))))


class _PerspectiveWidgetMessage(object):
    '''A custom message that will be passed from the Python widget to the
    front-end.

    When creating new messages, use this class as it defines a concrete schema
    for the message and prevents loosely creating `dict` objects everywhere.
    Use `to_dict()` to obtain the message in a form that can be sent through
    IPyWidgets.
    '''

    def __init__(self, msg_id, msg_type, msg_data):
        '''Create a new PerspectiveWidgetMessage.'''
        self.id = msg_id
        self.type = msg_type
        self.data = msg_data

    def to_dict(self):
        '''Returns a dictionary representation of the message.'''
        return {
            "id": self.id,
            "type": self.type,
            "data": self.data
        }


class PerspectiveWidget(Widget, PerspectiveViewer):
    ''':class`~perspective.PerspectiveWidget` allows for Perspective to be used
    in the form of a JupyterLab IPython widget.

    Using `perspective.Table`, you can create a widget that extends the full
    functionality of `perspective-viewer`.  Changes on the viewer can be
    programatically set on the :class`~perspective.PerspectiveWidget` instance,
    and state is maintained across page refreshes.

    Examples:
        >>> from perspective import Table, PerspectiveWidget
        >>> data = {
        ...     "a": [1, 2, 3],
        ...     "b": [
        ...         "2019/07/11 7:30PM",
        ...         "2019/07/11 8:30PM",
        ...         "2019/07/11 9:30PM"
        ...     ]
        ... }
        >>> tbl = Table(data, index="a")
        >>> widget = PerspectiveWidget(
        ...     tbl,
        ...     row_pivots=["a"],
        ...     sort=[["b", "desc"]],
        ...     filter=[["a", ">", 1]]
        ... )
        >>> widget.sort
        [["b", "desc"]]
        >>> widget.sort.append(["a", "asc"])
        >>> widget.sort
        [["b", "desc"], ["a", "asc"]]
        >>> widget.update({"a": [4, 5]}) # Browser UI updates
    '''

    # Required by ipywidgets for proper registration of the backend
    _model_name = Unicode('PerspectiveModel').tag(sync=True)
    _model_module = Unicode('@finos/perspective-jupyterlab').tag(sync=True)
    _model_module_version = Unicode("~{}".format(major_minor_version)).tag(sync=True)
    _view_name = Unicode('PerspectiveView').tag(sync=True)
    _view_module = Unicode('@finos/perspective-jupyterlab').tag(sync=True)
    _view_module_version = Unicode("~{}".format(major_minor_version)).tag(sync=True)

    def __init__(self,
                 table_or_data,
                 index=None,
                 limit=None,
                 client=not is_libpsp(),
                 **kwargs):
        '''Initialize an instance of :class`~perspective.PerspectiveWidget`
        with the given table/data and viewer configuration.

        If a pivoted DataFrame or MultiIndex table is passed in, the widget
        preserves pivots and applies them.  See `PerspectiveViewer.__init__` for
        arguments that transform the view shown in the widget.

        Args:
            table_or_data (perspective.Table|dict|list|pandas.DataFrame): The
                `Table` or data that will be viewed in the widget.

        Keyword Arguments:
            index (`str`): A column name to be used as the primary key.
                Ignored if a `Table` is supplied.
            limit (`int`): A upper limit on the number of rows in the Table.
                Cannot be set at the same time as `index`, ignored if a `Table`
                is passed in.

            client (`bool`):  If True, convert the dataset into an Apache Arrow
                binary and create the Table in Javascript using a copy of the
                data. Defaults to False.

            kwargs (`dict`): configuration options for the `PerspectiveViewer`,
                and `Table` constructor if `table_or_data` is a dataset.

        Examples:
            >>> widget = PerspectiveWidget(
            ...     {"a": [1, 2, 3]},
            ...     aggregates={"a": "avg"},
            ...     row_pivots=["a"],
            ...     sort=[["b", "desc"]],
            ...     filter=[["a", ">", 1]])
        '''
        self._displayed = False
        self.on_displayed(self._on_display)

        # If `self.client` is True, the front-end `perspective-viewer` is given
        # a copy of the data serialized to Arrow, and changes made in Python
        # do not reflect to the front-end.
        self.client = client

        if self.client:
            # Pass table load options to the front-end in client mode
            self._client_options = {}

            # Cache calls to `update()` before the widget has been displayed.
            self._predisplay_update_cache = []

        if index is not None and limit is not None:
            raise PerspectiveError("Index and Limit cannot be set at the same time!")

        # Parse the dataset we pass in - if it's Pandas, preserve pivots
        if isinstance(table_or_data, pandas.DataFrame) or isinstance(table_or_data, pandas.Series):
            data, config = deconstruct_pandas(table_or_data)
            table_or_data = data

            if config.get("row_pivots", None) and "row_pivots" not in kwargs:
                kwargs.update({"row_pivots": config["row_pivots"]})

            if config.get("column_pivots", None) and "column_pivots" not in kwargs:
                kwargs.update({"column_pivots": config["column_pivots"]})

            if config.get("columns", None) and "columns" not in kwargs:
                kwargs.update({"columns": config["columns"]})

        # Initialize the viewer
        super(PerspectiveWidget, self).__init__(**kwargs)

        # Handle messages from the the front end
        # `PerspectiveJupyterClient.send()`:
        # - The "data" value of the message should be a JSON-serialized string.
        # - Both `on_msg` and `@observe("value")` must be specified on the
        # handler for custom messages to be parsed by the Python widget.
        self.on_msg(self.handle_message)

        if self.client:
            if is_libpsp():
                from ..libpsp import Table
                if isinstance(table_or_data, Table):
                    raise PerspectiveError("Client mode PerspectiveWidget expects data or schema, not a `perspective.Table`!")

            if index is not None:
                self._client_options["index"] = index

            if limit is not None:
                self._client_options["limit"] = limit

            # cache self._data so creating multiple views don't reserialize the
            # same data
            if not hasattr(self, "_data") or self._data is None:
                self._data = _serialize(table_or_data)
        else:
            #  If an empty dataset is provided, don't call `load()`
            load_kwargs = {}
            if table_or_data is None:
                if index is not None or limit is not None:
                    raise PerspectiveError("Cannot initialize PerspectiveWidget `index` or `limit` without a Table, data, or schema!")
            else:
                if index is not None:
                    load_kwargs.update({"index": index})

                if limit is not None:
                    load_kwargs.update({"limit": limit})
                self.load(table_or_data, **load_kwargs)

    def load(self, data, **options):
        '''Load the widget with data. If running in client mode, this method
        serializes the data and calls the browser viewer's load method.
        Otherwise, it calls `Viewer.load()` using `super()`.
        '''
        if self.client is True:
            # serialize the data and send a custom message to the browser
            if isinstance(data, pandas.DataFrame) or isinstance(data, pandas.Series):
                data, _ = deconstruct_pandas(data)
            d = _serialize(data)
            self._data = d
        else:
            super(PerspectiveWidget, self).load(data, **options)

        # proactively notify front-end of new data
        message = self._make_load_message()
        self.send(message.to_dict())

    def update(self, data):
        '''Update the widget with new data. If running in client mode, this
        method serializes the data and calls the browser viewer's update
        method. Otherwise, it calls `Viewer.update()` using `super()`.
        '''
        if self.client is True:
            if self._displayed is False:
                self._predisplay_update_cache.append(data)
                return

            # serialize the data and send a custom message to the browser
            if isinstance(data, pandas.DataFrame) or isinstance(data, pandas.Series):
                data, _ = deconstruct_pandas(data)
            d = _serialize(data)
            self.post({
                "cmd": "update",
                "data": d
            })
        else:
            super(PerspectiveWidget, self).update(data)

    def clear(self):
        '''Clears the widget's underlying `Table`.

        In client mode, clears the `_data` attribute of the widget.
        '''
        if self.client is True:
            self.post({
                "cmd": "clear"
            })
            self._data = None
        else:
            super(PerspectiveWidget, self).clear()

    def replace(self, data):
        '''Replaces the widget's `Table` with new data conforming to the same
        schema. Does not clear user-set state. If in client mode, serializes
        the data and sends it to the browser.
        '''
        if self.client is True:
            if isinstance(data, pandas.DataFrame) or isinstance(data, pandas.Series):
                data, _ = deconstruct_pandas(data)
            d = _serialize(data)
            self.post({
                "cmd": "replace",
                "data": d
            })
            self._data = d
        else:
            super(PerspectiveWidget, self).replace(data)

    def delete(self, delete_table=True):
        '''Delete the Widget's data and clears its internal state. If running in
        client mode, sends the `delete()` command to the browser. Otherwise
        calls `delete` on the underlying viewer.

        Args:
            delete_table (`bool`): whether the underlying `Table` will be
                deleted. Defaults to True.
        '''
        if self.client is False:
            super(PerspectiveWidget, self).delete(delete_table)
        self.post({
            "cmd": "delete"
        })

        # Close the underlying comm and remove widget from the front-end
        self.close()

    def post(self, msg, msg_id=None):
        '''Post a serialized message to the `PerspectiveJupyterClient`
        in the front end.

        The posted message should conform to the `PerspectiveJupyterMessage`
        interface as defined in `@finos/perspective-jupyterlab`.

        Args:
            msg (dict): a message from `PerspectiveManager` for the front-end
                viewer to process.
            msg_id (int): an integer id that allows the client to process
                the message.
        '''
        message = _PerspectiveWidgetMessage(msg_id, "cmd", msg)
        self.send(message.to_dict())

    @observe("value")
    def handle_message(self, widget, content, buffers):
        '''Given a message from `PerspectiveJupyterClient.send()`, process the
        message and return the result to `self.post`.

        Args:
            widget: a reference to the `Widget` instance that received the
                message.
            content (dict): the message from the front-end. Automatically
                de-serialized by ipywidgets.
            buffers : optional arraybuffers from the front-end, if any.
        '''
        if content["type"] == "cmd":
            parsed = json.loads(content["data"])

            if parsed["cmd"] == "init":
                self.post({'id': -1, 'data': None})
            elif parsed["cmd"] == "table":
                # return the dataset or table name to the front-end
                msg = self._make_load_message()
                self.send(msg.to_dict())

                # In client mode, users can call `update()` before the widget
                # is visible. This applies the updates after the viewer has
                # loaded the initial dataset.
                if self.client is True and len(self._predisplay_update_cache) > 0:
                    for data in self._predisplay_update_cache:
                        self.update(data)
            else:
                # For all calls to Perspective, process it in the manager.
                post_callback = partial(self.post, msg_id=parsed["id"])
                self.manager._process(parsed, post_callback)

    def _make_load_message(self):
        '''Send a message to the front-end either containing the name of a
        Table in python, or the serialized dataset with options while in client
        mode.
        '''
        msg_data = None
        if self.client and self._data is not None:
            # Send data to the client, transferring ownership to the browser
            msg_data = {
                "data": self._data
            }

            if len(self._client_options.keys()) > 0:
                msg_data["options"] = self._client_options
        elif self.table_name is not None:
            # Only pass back the table if it's been loaded. If the table isn't
            # loaded, the `load()` method will handle synchronizing the
            # front-end.
            msg_data = {
                "table_name": self.table_name
            }

        if msg_data is not None:
            return _PerspectiveWidgetMessage(-2, "table", msg_data)
        else:
            raise PerspectiveError("Widget could not find a dataset or a `Table` to load.")

    def _on_display(self, widget, **kwargs):
        '''When the widget has been displayed, make sure `displayed` is set to
        True so updates stop being cached.
        '''
        self._displayed = True

    def on_update(self, callback, remove=False):
        self.manager.on_update(callback, remove=remove)



import math
import tributary.streaming as ts


def foo():
    yield 1
    yield 2
    yield 3
    yield 4
    yield 5


def foofloat():
    yield 1.11
    yield 2.22
    yield 3.33
    yield 4.44
    yield 5.55


def foo2():
    yield 1
    yield 4
    yield 9


def foo3():
    yield -1
    yield -2


def foo4():
    yield [1]
    yield [1, 2]
    yield [1, 2, 3]


class TestOps:
    def test_Noop(self):
        t = ts.Timer(foo, count=2)
        out = ts.Noop(t)
        assert ts.run(out) == [1, 2]

    def test_Negate(self):
        t = ts.Timer(foo, count=2)
        out = ts.Negate(t)
        assert ts.run(out) == [-1, -2]

    def test_Invert(self):
        t = ts.Timer(foo, count=2)
        out = ts.Invert(t)
        assert ts.run(out) == [1 / 1, 1 / 2]

    def test_Add(self):
        t = ts.Timer(foo, count=2)
        out = ts.Add(t, t)
        assert ts.run(out) == [2, 4]

    def test_Sub(self):
        t = ts.Timer(foo, count=2)
        out = ts.Sub(t, t)
        assert ts.run(out) == [0, 0]

    def test_Mult(self):
        t = ts.Timer(foo, count=2)
        out = ts.Mult(t, t)
        assert ts.run(out) == [1, 4]

    def test_Div(self):
        t = ts.Timer(foo, count=2)
        out = ts.Div(t, t)
        assert ts.run(out) == [1, 1]

    def test_RDiv(self):
        t = ts.Timer(foo, count=2)
        out = ts.RDiv(t, t)
        assert ts.run(out) == [1, 1]

    def test_Mod(self):
        t = ts.Timer(foo, count=5)
        c = ts.Const(3)
        out = ts.Mod(t, c)
        assert ts.run(out) == [1, 2, 0, 1, 2]

    def test_Pow(self):
        t = ts.Timer(foo, count=2)
        c = ts.Const(2)
        out = ts.Pow(t, c)
        assert ts.run(out) == [1, 4]

    def test_Sum(self):
        t = ts.Timer(foo, count=2)
        t2 = ts.Timer(foo, count=2)
        c = ts.Const(2)
        out = ts.Sum(t, t2, c)
        assert ts.run(out) == [4, 6]

    def test_Average(self):
        t = ts.Timer(foo, count=1)
        t2 = ts.Timer(foo, count=1)
        c = ts.Const(1)
        out = ts.Average(t, t2, c)
        assert ts.run(out) == [1]

    def test_Not(self):
        t = ts.Timer(foo, count=2)
        out = ts.Not(t)
        assert ts.run(out) == [False, False]

    def test_And(self):
        t = ts.Timer(foo, count=2)
        out = ts.And(t, t)
        assert ts.run(out) == [1, 2]

    def test_Or(self):
        t = ts.Timer(foo, count=2)
        out = ts.Or(t, t)
        assert ts.run(out) == [1, 2]

    def test_Equal(self):
        t = ts.Timer(foo, count=2)
        c = ts.Const(1)
        out = ts.Equal(t, c)
        assert ts.run(out) == [True, False]

    def test_NotEqual(self):
        t = ts.Timer(foo, count=2)
        c = ts.Const(1)
        out = ts.NotEqual(t, c)
        assert ts.run(out) == [False, True]

    def test_Lt(self):
        t = ts.Timer(foo, count=2)
        c = ts.Const(1)
        out = ts.Lt(c, t)
        assert ts.run(out) == [False, True]

    def test_Le(self):
        t = ts.Timer(foo, count=2)
        c = ts.Const(2)
        out = ts.Le(c, t)
        assert ts.run(out) == [False, True]

    def test_Gt(self):
        t = ts.Timer(foo, count=2)
        c = ts.Const(1)
        out = ts.Gt(t, c)
        assert ts.run(out) == [False, True]

    def test_Ge(self):
        t = ts.Timer(foo, count=2)
        c = ts.Const(2)
        out = ts.Ge(t, c)
        assert ts.run(out) == [False, True]

    def test_Log(self):
        t = ts.Timer(foo, count=2)
        out = ts.Log(t)
        assert ts.run(out) == [math.log(1), math.log(2)]

    def test_Sin(self):
        t = ts.Timer(foo, count=2)
        out = ts.Sin(t)
        assert ts.run(out) == [math.sin(1), math.sin(2)]

    def test_Cos(self):
        t = ts.Timer(foo, count=2)
        out = ts.Cos(t)
        assert ts.run(out) == [math.cos(1), math.cos(2)]

    def test_Tan(self):
        t = ts.Timer(foo, count=2)
        out = ts.Tan(t)
        assert ts.run(out) == [math.tan(1), math.tan(2)]

    def test_Arcsin(self):
        t = ts.Timer(foo, count=2)
        c = ts.Const(value=3)
        out = ts.Arcsin(ts.Div(t, c))
        assert ts.run(out) == [math.asin(1 / 3), math.asin(2 / 3)]

    def test_Arccos(self):
        t = ts.Timer(foo, count=2)
        c = ts.Const(value=3)
        out = ts.Arccos(ts.Div(t, c))
        assert ts.run(out) == [math.acos(1 / 3), math.acos(2 / 3)]

    def test_Arctan(self):
        t = ts.Timer(foo, count=2)
        out = ts.Arctan(t)
        assert ts.run(out) == [math.atan(1), math.atan(2)]

    def test_Sqrt(self):
        t = ts.Timer(foo2, count=3)
        out = ts.Sqrt(t)
        assert ts.run(out) == [1.0, 2.0, 3.0]

    def test_Abs(self):
        t = ts.Timer(foo3, count=2)
        out = ts.Abs(t)
        assert ts.run(out) == [1, 2]

    def test_Exp(self):
        t = ts.Timer(foo, count=2)
        out = ts.Exp(t)
        assert ts.run(out) == [math.exp(1), math.exp(2)]

    def test_Erf(self):
        t = ts.Timer(foo, count=2)
        out = ts.Erf(t)
        assert ts.run(out) == [math.erf(1), math.erf(2)]

    def test_Floor(self):
        t = ts.Timer(foofloat, count=2)
        out = ts.Floor(t)
        assert ts.run(out) == [1., 2.]

    def test_Ceil(self):
        t = ts.Timer(foofloat, count=2)
        out = ts.Ceil(t)
        assert ts.run(out) == [2., 3.]

    def test_Round(self):
        t = ts.Timer(foofloat, count=2)
        out = ts.Round(t, 1)
        assert ts.run(out) == [1.1, 2.2]

    def test_Int(self):
        t = ts.Timer(foo, count=2)
        out = ts.Int(t)
        assert ts.run(out) == [1, 2]

    def test_Float(self):
        t = ts.Timer(foo, count=2)
        out = ts.Float(t)
        assert ts.run(out) == [1.0, 2.0]

    def test_Bool(self):
        t = ts.Timer(foo, count=2)
        out = ts.Bool(t)
        assert ts.run(out) == [True, True]

    def test_Str(self):
        t = ts.Timer(foo, count=2)
        out = ts.Str(t)
        assert ts.run(out) == ['1', '2']

    def test_Len(self):
        t = ts.Timer(foo4, count=3)
        out = ts.Len(t)
        assert ts.run(out) == [1, 2, 3]



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

BOOLEAN_FILTERS = ["&", "|", "==", "!=", "or", "and"]
NUMBER_FILTERS = ["<", ">", "==", "<=", ">=", "!=", "is null", "is not null"]
STRING_FILTERS = ["==", "contains", "!=", "in", "not in", "begins with", "ends with"]
DATETIME_FILTERS = ["<", ">", "==", "<=", ">=", "!="]

ALL_FILTERS = BOOLEAN_FILTERS + NUMBER_FILTERS + STRING_FILTERS + DATETIME_FILTERS



import json as JSON
from confluent_kafka import Consumer, KafkaError
from .input import Foo


class Kafka(Foo):
    '''Connect to kafka server and yield back results

    Args:
        servers (list): kafka bootstrap servers
        group (str): kafka group id
        topics (list): list of kafka topics to connect to
        json (bool): load input data as json
        wrap (bool): wrap result in a list
        interval (int): kafka poll interval
    '''

    def __init__(self, servers, group, topics, json=False, wrap=False, interval=1):
        c = Consumer({
            'bootstrap.servers': servers,
            'group.id': group,
            'default.topic.config': {
                'auto.offset.reset': 'smallest'
            }
        })

        if not isinstance(topics, list):
            topics = [topics]
        c.subscribe(topics)

        async def _listen(consumer=c, json=json, wrap=wrap, interval=interval):
            while True:
                msg = consumer.poll(interval)

                if msg is None:
                    continue
                if msg.error():
                    if msg.error().code() == KafkaError._PARTITION_EOF:
                        continue
                    else:
                        print(msg.error())
                        break

                msg = msg.value().decode('utf-8')

                if not msg:
                    break
                if json:
                    msg = JSON.loads(msg)
                if wrap:
                    msg = [msg]
                yield msg

        super().__init__(foo=_listen)
        self._name = 'Kafka'



import time
import tributary.functional as t


class TestInput:
    def test_http(self):
        t.pipeline([t.http], ['callback'], [{'url': 'https://www.google.com'}], on_data=print)
        time.sleep(2)
        t.stop()



import asyncio
import tributary.streaming as ts


class TestStreaming:
    def test_run_simple(self):
        t = ts.Const(value=1, count=1)
        assert ts.run(t) == [1]

    def test_run_foo(self):
        def foo():
            return 5
        t = ts.Foo(foo, count=1)
        assert ts.run(t) == [5]

    def test_run_generator(self):
        def foo():
            yield 1
            yield 2
        t = ts.Foo(foo)
        assert ts.run(t) == [1, 2]

    def test_run_async_foo(self):
        async def foo():
            await asyncio.sleep(.1)
            return 5

        t = ts.Foo(foo, count=1)
        assert ts.run(t) == [5]

    def test_run_async_generator(self):
        async def foo():
            yield 1
            yield 2
        t = ts.Foo(foo)
        assert ts.run(t) == [1, 2]



# *****************************************************************************
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
import six
from perspective.table import Table
from datetime import date, datetime


class TestTableInfer(object):

    def test_table_infer_int(self):
        data = {"a": [None, None, None, None, 1, 0, 1, 1, 1]}
        tbl = Table(data)
        assert tbl.schema() == {"a": int}

    def test_table_infer_float(self):
        data = {"a": [None, None, None, None, 1.0, 2.0]}
        tbl = Table(data)
        assert tbl.schema() == {"a": float}

    def test_table_infer_bool(self):
        bool_data = [{"a": True, "b": False}, {"a": True, "b": True}]
        tbl = Table(bool_data)
        assert tbl.size() == 2
        assert tbl.schema() == {
            "a": bool,
            "b": bool
        }

    def test_table_infer_bool_str(self):
        bool_data = [{"a": "True", "b": "False"}, {"a": "True", "b": "True"}]
        tbl = Table(bool_data)
        assert tbl.size() == 2
        assert tbl.schema() == {
            "a": bool,
            "b": bool
        }

    def test_table_bool_infer_str_all_formats_from_schema(self):
        bool_data = [
            {"a": "True", "b": "False"},
            {"a": "t", "b": "f"},
            {"a": "true", "b": "false"},
            {"a": 1, "b": 0},
            {"a": "on", "b": "off"}
        ]
        tbl = Table(bool_data)
        assert tbl.schema() == {
            "a": bool,
            "b": bool
        }
        assert tbl.size() == 5
        assert tbl.view().to_dict() == {
            "a": [True, True, True, True, True],
            "b": [False, False, False, False, False]
        }

    def test_table_promote_float(self):
        if six.PY2:
            data = {"a": [1.5, 2.5, 3.5, 4.5, "abc"]}
            tbl = Table(data)
            assert tbl.schema() == {"a": str}
            assert tbl.view().to_dict() == {"a": ["1.5", "2.5", "3.5", "4.5", "abc"]}

    def test_table_promote_float_py2(self):
        if six.PY2:
            data = {"a": [1, 2, 3, 4, 2147483648]}
            tbl = Table(data)
            assert tbl.schema() == {"a": float}
            assert tbl.view().to_dict() == {"a": [1.0, 2.0, 3.0, 4.0, 2147483648.0]}

    def test_table_infer_bool(self):
        data = {"a": [None, None, None, None, True, True, True]}
        tbl = Table(data)
        assert tbl.schema() == {"a": bool}

    def test_table_infer_str(self):
        data = {"a": [None, None, None, None, None, None, "abc"]}
        tbl = Table(data)
        assert tbl.schema() == {"a": str}

    def test_table_infer_time_as_string(self):
        # time objects are inferred as string
        data = {"a": [None, None, None, None, None, None, datetime(2019, 7, 11, 12, 30, 5).time()]}
        tbl = Table(data)
        assert tbl.schema() == {"a": str}

    def test_table_infer_date_from_datetime(self):
        # inferrence on non-pandas datasets defaults to datetime
        data = {"a": [None, None, None, None, None, None, datetime(2019, 7, 11)]}
        tbl = Table(data)
        assert tbl.schema() == {"a": datetime}

    def test_table_infer_date_from_date(self):
        # pass in a `date` to make sure it infers as date
        data = {"a": [None, None, None, None, None, None, date(2019, 7, 11)]}
        tbl = Table(data)
        assert tbl.schema() == {"a": date}

    def test_table_infer_valid_date(self):
        data = {"a": [None, None, None, None, None, None, "08/31/2019"]}
        tbl = Table(data)
        assert tbl.schema() == {"a": date}

    def test_table_infer_ambiguous_date(self):
        data = {"a": [None, None, None, None, None, None, "01/03/2019"]}
        tbl = Table(data)
        assert tbl.schema() == {"a": date}

    def test_table_infer_ymd_date(self):
        data = {"a": [None, None, None, None, None, None, "2019/01/03"]}
        tbl = Table(data)
        assert tbl.schema() == {"a": date}

    def test_table_infer_invalid_date(self):
        data = {"a": [None, None, None, None, None, None, "08/55/2019"]}
        tbl = Table(data)
        assert tbl.schema() == {"a": str}

    def test_table_infer_date_edge(self):
        data = {"a": [None, None, None, None, None, None, "08/31/2019 00:00:00"]}
        tbl = Table(data)
        assert tbl.schema() == {"a": date}

    def test_table_infer_datetime_edge(self):
        data = {"a": [None, None, None, None, None, None, "08/31/2019 00:00:01"]}
        tbl = Table(data)
        assert tbl.schema() == {"a": datetime}

    def test_table_infer_valid_datetime(self):
        data = {"a": [None, None, None, None, None, None, "08/31/2019 07:30:00"]}
        tbl = Table(data)
        assert tbl.schema() == {"a": datetime}

    def test_table_infer_iso_datetime(self):
        data = {"a": [None, None, None, None, None, None, "2019/07/25T09:00:00"]}
        tbl = Table(data)
        assert tbl.schema() == {"a": datetime}

    def test_table_infer_datetime_separators(self):
        data = {"a": [None, None, None, None, None, "2019-07-25T09:00:00", "2019/07/25T09:00:00"]}
        tbl = Table(data)
        assert tbl.schema() == {"a": datetime}

    def test_table_infer_datetime_tz(self):
        data = {"a": [None, None, None, None, None, "2019-07-25T09:00:00-05:00"]}
        tbl = Table(data)
        assert tbl.schema() == {"a": datetime}

    def test_table_infer_invalid_datetime(self):
        data = {"a": [None, None, None, None, None, None, "08/31/2019 25:30:00"]}
        tbl = Table(data)
        assert tbl.schema() == {"a": str}

    def test_table_infer_mixed_date(self):
        data = {"a": [None, None, None, None, None, "08/11/2019"]}
        tbl = Table(data)
        assert tbl.schema() == {"a": date}

    def test_table_infer_mixed_datetime(self):
        data = {"a": [None, None, None, None, None, "08/11/2019 13:14:15"]}
        tbl = Table(data)
        assert tbl.schema() == {"a": datetime}

    def test_table_strict_datetime_infer(self):
        data = {"a": ['10', '9', '8', '7', '6', '5', '4', '3', '2', '1']}
        tbl = Table(data)
        assert tbl.schema() == {"a": str}

    def test_table_strict_date_infer(self):
        data = {"a": ["2019 09 10"]}
        tbl = Table(data)
        assert tbl.schema() == {"a": date}

    def test_table_strict_datetime_separator_infer(self):
        data = {"a": ["2019-10-01 7:30"]}
        tbl = Table(data)
        assert tbl.schema() == {"a": datetime}

    def test_table_datetime_infer_no_false_positive(self):
        data = {"a": [" . - / but clearly not a date"]}
        tbl = Table(data)
        assert tbl.schema() == {"a": str}



import string
import itertools
import numpy as np
import pandas as pd
from functools import lru_cache
from tributary.lazy import Node
from ipywidgets import Output, VBox
from perspective import Table, PerspectiveWidget


@lru_cache(None)
def _genData():
    return list(string.ascii_uppercase)  # + list(''.join(_) for _ in itertools.product(string.ascii_uppercase, repeat=2))


class Grid(object):
    def __init__(self):
        schema = {x: str for x in _genData()}
        schema['index'] = int

        self._table = Table(schema, index='index')
        l = np.empty((100, ), str)
        data = {x: l for x in _genData()}
        data['index'] = np.arange(100)
        self._table.update(pd.DataFrame(data))
        self._widget = PerspectiveWidget(self._table, columns=_genData(), editable=True)
        self._widget.on_update(self.handle_message)
        self._out = Output()
        self._graph = VBox([])
        self._nodes = {}

    def render(self):
        return VBox([self._widget, self._out, self._graph])

    def handle_message(self, widget, content):
        with self._out:
            data = content['args'][0]
            for update in data:
                self._process(update)

    def _process(self, data):
        keys = list(set(data.keys()) - set(("__INDEX__",)))
        out_symbol = keys[0] + str(data['__INDEX__'][0])
        print(out_symbol)

        data = data[keys[0]]
        print(data)

        if out_symbol not in self._nodes:
            self._nodes[out_symbol] = Node(value=data)
        else:
            self._nodes[out_symbol].setValue(data)

        self._graph.children = [self._nodes[out_symbol].dagre()]



import tributary.lazy as t
import random


class Foo1(t.LazyGraph):
    def __init__(self, *args, **kwargs):
        self.x = self.node('x', readonly=False, value=1, trace=True)


class Foo2(t.LazyGraph):
    def __init__(self, *args, **kwargs):
        self.y = self.node('y', readonly=False, value=2, trace=True)

        # ensure no __nodes clobber
        self.test = self.node('test', readonly=False, value=2, trace=True)
        self.x = self.node('x', readonly=False, value=2, trace=True)


class Foo3(t.LazyGraph):
    @t.node(trace=True)
    def foo1(self):
        return self.random()  # test self access

    def random(self):
        return random.random()

    @t.node(trace=True)
    def foo3(self, x=4):
        return 3 + x


class Foo4(t.LazyGraph):
    @t.node(trace=True)
    def foo1(self):
        return self.foo2() + 1

    @t.node(trace=True)
    def foo2(self):
        return random.random()


class Foo5(t.LazyGraph):
    @t.node()
    def z(self):
        return self.x | self.y()

    @t.node()
    def y(self):
        return 10

    def reset(self):
        self.x = None

    def __init__(self):
        self.x = self.node(name="x", value=None)


class TestLazy:
    def test_misc(self):
        f4 = Foo4()
        z = f4.foo1()
        assert z.print()
        assert z.graph()
        assert z.graphviz()


class TestDirtyPropogation:
    def test_or_dirtypropogation(self):
        f = Foo5()
        assert f.z()() == 10
        assert f.x() is None

        f.x = 5

        assert f.x() == 5
        assert f.z()() == 5

        f.reset()

        assert f.x() is None
        assert f.z()() == 10


class TestDeclarative:
    def test_simple_declarative(self):
        n = t.Node(value=1)
        z = n + 5
        assert z() == 6
        n.setValue(2)
        assert z() == 7



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#


class _PerspectiveCallBackCache(object):
    def __init__(self):
        self._callbacks = []

    def add_callback(self, callback):
        self._callbacks.append(callback)

    def remove_callbacks(self, condition):
        '''Remove callback functions that satisfy the given condition.

        Args:
            condition (func): a function that returns either True or False. If
                True is returned, filter the item out.
        '''
        if not callable(condition):
            raise ValueError("callback filter condition must be a callable function!")
        self._callbacks = [callback for callback in self._callbacks if condition(callback) is True]

    def get_callbacks(self):
        return self._callbacks

    def __repr__(self):
        return str(self._callbacks)



from __future__ import print_function

import concurrent.futures.thread as cft
from concurrent.futures import ThreadPoolExecutor, _base
from concurrent.futures.thread import _WorkItem
from functools import partial
from .input import *  # noqa: F401, F403
from .utils import *  # noqa: F401, F403

try:
    # For Travis/ python 3.6
    from concurrent.futures.thread import BrokenThreadPool
except ImportError:
    BrokenThreadPool = Exception


_EXECUTOR = ThreadPoolExecutor(max_workers=10)


def submit(fn, *args, **kwargs):
    '''Submit a function to be run on the executor (internal)

    Args:
        fn (callable): function to call
        args (tuple): args to pass to function
        kwargs (dict): kwargs to pass to function
    '''
    if _EXECUTOR is None:
        raise RuntimeError('Already stopped!')
    self = _EXECUTOR
    with self._shutdown_lock:
        if hasattr(self, '_broken') and self._broken:
            raise BrokenThreadPool(self._broken)

        if hasattr(self, '_shutdown') and self._shutdown:
            raise RuntimeError('cannot schedule new futures after shutdown')
        if cft._shutdown:
            raise RuntimeError('cannot schedule new futures after'
                               'interpreter shutdown')

        f = _base.Future()
        w = _WorkItem(f, fn, args, kwargs)

        self._work_queue.put(w)
        self._adjust_thread_count()
        return f


def run_submit(fn, function_to_call, *args, **kwargs):
    try:
        f = submit(fn, *args, **kwargs)
    except RuntimeError:
        # means we've shutdown, stop
        return

    if function_to_call:
        f.add_done_callback(lambda fut: function_to_call(fut.result()) if fut.result() else None)


def pipeline(foos, foo_callbacks, foo_kwargs=None, on_data=print, on_data_kwargs=None):
    '''Pipeline a sequence of functions together via callbacks

    Args:
        foos (list of callables): list of functions to pipeline
        foo_callbacks (List[str]): list of strings indicating the callback names (kwargs of the foos)
        foo_kwargs (List[dict]):
        on_data (callable): callable to call at the end of the pipeline
        on_data_kwargs (dict): kwargs to pass to the on_data function>?
    '''
    global _EXECUTOR
    if _EXECUTOR is None:
        _EXECUTOR = ThreadPoolExecutor(max_workers=2)

    foo_kwargs = foo_kwargs or []
    on_data_kwargs = on_data_kwargs or {}

    # organize args for functional pipeline
    assembled = []
    for i, foo in enumerate(foos):
        cb = foo_callbacks[i] if i < len(foo_callbacks) else 'on_data'
        kwargs = foo_kwargs[i] if i < len(foo_kwargs) else {}
        assembled.append((foo, cb, kwargs))

    # assemble pipeline
    assembled.reverse()
    lambdas = [lambda d, f=on_data: run_submit(f, None, d, **on_data_kwargs)]
    for i, a in enumerate(assembled):
        foo, cb, kwargs = a
        function_to_call = lambdas[i]
        kwargs[cb] = function_to_call

        if i != len(assembled) - 1:
            lambdas.append(lambda d, kw=kwargs, f=foo: run_submit(f, function_to_call, d, **kw))
            lambdas[-1].__name__ = foo.__name__
        else:
            lambdas.append(lambda kw=kwargs, f=foo: run_submit(f, function_to_call, **kw))
            lambdas[-1].__name__ = foo.__name__

    # start entrypoint
    lambdas[-1]()


def stop():
    '''Stop the executor for the pipeline runtime'''
    global _EXECUTOR
    _EXECUTOR.shutdown(False)
    _EXECUTOR._threads.clear()
    cft._threads_queues.clear()
    _EXECUTOR = None


def wrap(function, *args, **kwargs):
    '''wrap a function in a partial'''
    foo = partial(function, *args, **kwargs)
    foo.__name__ = function.__name__
    return foo



from .node import Node, node  # noqa: F401


class LazyGraph(object):
    '''Wrapper class around a collection of lazy nodes.'''

    def __init__(self, *args, **kwargs):
        pass

    def node(self, name, readonly=False, nullable=True, value=None, trace=False):  # noqa: F811
        '''method to create a lazy node attached to a graph.

        Args:
            name (str): name to represent the node
            readonly (bool): whether the node should be settable
            nullable (bool): whether node can have value None
            value (any): initial value for node
            trace (bool): trace the calls for a node
        Returns:
            BaseNode: the newly constructed lazy node
        '''
        if not hasattr(self, '_LazyGraph__nodes'):
            self.__nodes = {}

        if name not in self.__nodes:
            if not isinstance(value, Node):
                value = Node(name=name,
                             derived=False,
                             readonly=readonly,
                             nullable=nullable,
                             value=value,
                             trace=trace)
            self.__nodes[name] = value
            setattr(self, name, self.__nodes[name])
        return self.__nodes[name]

    def __getattribute__(self, name):
        if name == '_LazyGraph__nodes' or name == '__nodes':
            return super(LazyGraph, self).__getattribute__(name)
        elif hasattr(self, '_LazyGraph__nodes') and name in super(LazyGraph, self).__getattribute__('_LazyGraph__nodes'):
            return super(LazyGraph, self).__getattribute__('_LazyGraph__nodes')[name]
        else:
            return super(LazyGraph, self).__getattribute__(name)

    def __setattr__(self, name, value):
        if hasattr(self, '_LazyGraph__nodes') and name in super(LazyGraph, self).__getattribute__('_LazyGraph__nodes'):
            node = super(LazyGraph, self).__getattribute__('_LazyGraph__nodes')[name]
            if isinstance(value, Node) and node == value:
                return
            elif isinstance(value, Node):
                raise Exception('Cannot set to node')
            else:
                node._dirty = (node._value != value) or (node._value is not None and abs(node._value - value) > 10**-5)
                node._value = value
        else:
            super(LazyGraph, self).__setattr__(name, value)


def construct(dag):
    '''dag is representation of the underlying DAG'''



from ..node import Node


def _print(node, cache=None):
    if cache is None:
        cache = {}

    cache[id(node)] = node

    ret = {node: []}

    if node._dependencies:
        for call, deps in node._dependencies.items():
            # callable node
            if hasattr(call, '_node_wrapper') and \
                    call._node_wrapper is not None:
                val = call._node_wrapper._print(cache)
                ret[node].append(val)

            # args
            for arg in deps[0]:
                val = arg._print(cache)
                ret[node].append(val)

            # kwargs
            for kwarg in deps[1].values():
                val = kwarg._print(cache)
                ret[node].append(val)

    return ret


def Print(node):
    return node._print({})


def Graph(node):
    return node.print()


def GraphViz(node):
    d = node.graph()

    from graphviz import Digraph
    dot = Digraph(node._name, strict=True)
    dot.format = 'png'

    def rec(nodes, parent):
        for d in nodes:
            if not isinstance(d, dict):
                if d.isDirty():
                    dot.node(d._name, color='red', shape=d._graphvizshape)
                    dot.edge(d._name, parent._name, color='red')
                else:
                    dot.node(d._name, shape=d._graphvizshape)
                    dot.edge(d._name, parent._name)
            else:
                for k in d:
                    if k.isDirty():
                        dot.node(k._name, color='red', shape=k._graphvizshape)
                        rec(d[k], k)
                        dot.edge(k._name, parent._name, color='red')
                    else:
                        dot.node(k._name, shape=k._graphvizshape)
                        rec(d[k], k)
                        dot.edge(k._name, parent._name)

    for k in d:
        if k.isDirty():
            dot.node(k._name, color='red', shape=k._graphvizshape)
        else:
            dot.node(k._name, shape=k._graphvizshape)
        rec(d[k], k)
    return dot


def Dagre(node):
    import ipydagred3 as dd3
    G = dd3.Graph()
    d = Graph(node)

    def rec(nodes, parent):
        for d in nodes:
            if not isinstance(d, dict):
                d._dd3g = G
                if d.isDirty():
                    G.setNode(d._name, style='fill: #f00', shape="rect" if d._graphvizshape == "box" else d._graphvizshape)
                    # G.setEdge(d._name, parent._name, style='stroke: #f00')
                else:
                    G.setNode(d._name, style='fill: #fff', shape="rect" if d._graphvizshape == "box" else d._graphvizshape)

                G.setEdge(d._name, parent._name, style='stroke: #000')
            else:
                for k in d:
                    k._dd3g = G
                    if k.isDirty():
                        G.setNode(k._name, style='fill: #f00', shape="rect" if k._graphvizshape == "box" else k._graphvizshape)
                        rec(d[k], k)
                        # G.setEdge(k._name, parent._name, style='stroke: #f00')
                    else:
                        G.setNode(k._name, style='fill: #fff', shape="rect" if k._graphvizshape == "box" else k._graphvizshape)
                        rec(d[k], k)

                    G.setEdge(k._name, parent._name, style='stroke: #000')

    for k in d:
        k._dd3g = G
        if k.isDirty():
            G.setNode(k._name, style='fill: #f00', shape="rect" if k._graphvizshape == "box" else k._graphvizshape)
        else:
            G.setNode(k._name, style='fill: #fff', shape="rect" if k._graphvizshape == "box" else k._graphvizshape)
        rec(d[k], k)

    graph = dd3.DagreD3Widget(graph=G)
    return graph


Node._print = _print
Node.print = Print
Node.graph = Graph
Node.graphviz = GraphViz
Node.dagre = Dagre



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import json
import tornado.websocket
from tornado.ioloop import IOLoop
from ..core.exception import PerspectiveError


# Redefine `queue_process` to take advantage of `tornado.ioloop`
def _queue_process_tornado(table_id, state_manager):
    loop = IOLoop.current()
    loop.add_callback(state_manager.call_process, table_id=table_id)


class PerspectiveTornadoHandler(tornado.websocket.WebSocketHandler):
    '''PerspectiveTornadoHandler is a drop-in implementation of Perspective.

    Use it inside Tornado routing to create a server-side Perspective that is
    ready to receive websocket messages from the front-end `perspective-viewer`.
    Because Tornado implements an event loop, this handler links Perspective
    with `IOLoop.current()` in order to defer expensive operations until the
    next free iteration of the event loop.

    Examples:
        >>> MANAGER = PerspectiveManager()
        >>> MANAGER.host_table("data_source_one", Table(
        ...     pd.read_csv("superstore.csv")))
        >>> app = tornado.web.Application([
        ...     (r"/", MainHandler),
        ...     (r"/websocket", PerspectiveTornadoHandler, {
        ...         "manager": MANAGER,
        ...         "check_origin": True
        ...     })
        ... ])
    '''

    def __init__(self, *args, **kwargs):
        '''Create a new instance of the PerspectiveTornadoHandler with the
        given Manager instance.

        Keyword Args:
            manager (:obj`PerspectiveManager`): A `PerspectiveManager` instance.
                Must be provided on initialization.
            check_origin (:obj`bool`): If True, all requests will be accepted
                regardless of origin. Defaults to False.
        '''
        self._manager = kwargs.pop("manager", None)
        self._session = self._manager.new_session()
        self._check_origin = kwargs.pop("check_origin", False)

        if self._manager is None:
            raise PerspectiveError("A `PerspectiveManager` instance must be provided to the tornado handler!")

        super(PerspectiveTornadoHandler, self).__init__(*args, **kwargs)

        # make sure each `Table` calls the asynchronous version of `queue_process`
        self._manager._set_queue_process(_queue_process_tornado)

    def check_origin(self, origin):
        '''Returns whether the handler allows requests from origins outside
        of the host URL.
        '''
        return self._check_origin

    def on_message(self, message):
        '''When the websocket receives a message, send it to the `process`
        method of the `PerspectiveManager` with a reference to the `post`
        callback.
        '''
        if message == "heartbeat":
            return
        message = json.loads(message)
        self._session.process(message, self.post)

    def post(self, message, binary=False):
        '''When `post` is called by `PerspectiveManager`, serialize the data to
        JSON and send it to the client.

        Args:
            message (str): a JSON-serialized string containing a message to the
                front-end `perspective-viewer`.
        '''
        self.write_message(message, binary)

    def on_close(self):
        '''Remove the views associated with the client when the websocket
        closes.
        '''
        self._session.close()



import tributary.lazy as tl
from datetime import datetime
from time import sleep


class TestUtils:
    def test_expire(self):
        n = tl.Node(value=5)
        sec = datetime.now().second
        out = tl.Expire(n, second=(sec + 2) % 60)

        # assert initial value
        assert out() == 5

        # set new value
        n.setValue(6)

        # continue to use old value until 2+ seconds elapsed
        assert out() == 5

        sleep(3)
        assert out() == 6

    def test_interval(self):
        n = tl.Node(value=5)
        out = tl.Interval(n, seconds=2)

        # assert initial value
        assert out() == 5

        # set new value
        n.setValue(6)

        # continue to use old value until 2+ seconds elapsed
        assert out() == 5

        sleep(3)
        assert out() == 6



import pandas as pd
from random import random, randint, choice
from faker import Faker
fake = Faker()


def superstore(count=50):
    data = []
    for id in range(count):
        dat = {}
        dat['Row ID'] = id
        dat['Order ID'] = '{}-{}'.format(fake.ein(), fake.zipcode())
        dat['Order Date'] = fake.date_this_year()
        dat['Ship Date'] = fake.date_between_dates(dat['Order Date']).strftime('%Y-%m-%d')
        dat['Order Date'] = dat['Order Date'].strftime('%Y-%m-%d')
        dat['Ship Mode'] = choice(['First Class', 'Standard Class', 'Second Class'])
        dat['Ship Mode'] = choice(['First Class', 'Standard Class', 'Second Class'])
        dat['Customer ID'] = fake.zipcode()
        dat['Segment'] = choice(['A', 'B', 'C', 'D'])
        dat['Country'] = 'US'
        dat['City'] = fake.city()
        dat['State'] = fake.state()
        dat['Postal Code'] = fake.zipcode()
        dat['Region'] = choice(['Region %d' % i for i in range(5)])
        dat['Product ID'] = fake.bban()
        sector = choice(['Industrials', 'Technology', 'Financials'])
        industry = choice(['A', 'B', 'C'])
        dat['Category'] = sector
        dat['Sub-Category'] = industry
        dat['Sales'] = randint(1, 100) * 100
        dat['Quantity'] = randint(1, 100) * 10
        dat['Discount'] = round(random() * 100, 2)
        dat['Profit'] = round(random() * 1000, 2)
        data.append(dat)
    return pd.DataFrame(data)



################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

from six import iteritems, string_types
from ..core.exception import PerspectiveError
from ..core import Aggregate, Plugin, ALL_FILTERS, Sort


def validate_plugin(plugin):
    if isinstance(plugin, Plugin):
        return plugin.value
    elif isinstance(plugin, string_types):
        if plugin not in Plugin.options():
            raise PerspectiveError("Unrecognized `plugin`: {0}".format(plugin))
        return plugin
    else:
        raise PerspectiveError("Cannot parse `plugin` of type: {0}".format(str(type(plugin))))


def validate_columns(columns):
    if columns is None:
        return []
    elif isinstance(columns, string_types):
        columns = [columns]

    if isinstance(columns, list):
        return columns
    else:
        raise PerspectiveError('Cannot parse `columns` of type: %s', str(type(columns)))


def _validate_pivots(pivots):
    if pivots is None:
        return []
    elif isinstance(pivots, string_types):
        pivots = [pivots]

    if isinstance(pivots, list):
        return pivots
    else:
        raise PerspectiveError('Cannot parse pivots of type: %s', str(type(pivots)))


def validate_row_pivots(row_pivots):
    return _validate_pivots(row_pivots)


def validate_column_pivots(column_pivots):
    return _validate_pivots(column_pivots)


def validate_aggregates(aggregates):
    if aggregates is None:
        return {}
    elif isinstance(aggregates, dict):
        for k, v in iteritems(aggregates):
            if isinstance(v, Aggregate):
                aggregates[k] = v.value
            elif isinstance(v, string_types):
                if v not in Aggregate.options():
                    raise PerspectiveError('Unrecognized aggregate: %s', v)
            else:
                raise PerspectiveError('Cannot parse aggregation of type %s', str(type(v)))
        return aggregates
    else:
        raise PerspectiveError('Cannot parse aggregates type: %s', str(type(aggregates)))


def validate_sort(sort):
    if sort is None:
        return []
    elif isinstance(sort, string_types):
        sort = [sort]

    if isinstance(sort, list):
        if len(sort) > 0 and not isinstance(sort[0], list):
            sort = [sort]
        ret = []
        for col, s in sort:
            if isinstance(s, Sort):
                s = s.value
            elif not isinstance(s, string_types) or s not in Sort.options():
                raise PerspectiveError('Unrecognized sort direction: %s', s)
            ret.append([col, s])
        return ret
    else:
        raise PerspectiveError('Cannot parse sort type: %s', str(type(sort)))


def validate_filters(filters):
    if filters is None:
        return []

    elif isinstance(filters, list) and len(filters) > 0 and not isinstance(filters[0], list):
        # wrap
        filters = [filters]

    if isinstance(filters, list):
        for f in filters:
            if not isinstance(f, list):
                raise PerspectiveError('`filters` kwarg must be a list!')

            for i, item in enumerate(f):
                if i == 1:
                    if item not in ALL_FILTERS:
                        raise PerspectiveError('Unrecognized filter operator: {}'.format(item))
                    elif item not in ("is null", "is not null"):
                        if len(f) != 3:
                            raise PerspectiveError('Cannot parse filter - {} operator must have a comparison value.'.format(item))
        return filters
    else:
        raise PerspectiveError('Cannot parse filters type: {}'.format(str(type(filters))))


def validate_plugin_config(plugin_config):
    return plugin_config
